# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Mapping database models to REST Api support."""

import datetime as dt
from collections.abc import Callable
from copy import copy
from typing import Any, NamedTuple, Never, TYPE_CHECKING, TypeAlias, override

import pydantic
from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import QuerySet
from django.http import HttpRequest
from django.urls import reverse
from django.utils.encoding import smart_str
from rest_framework import serializers
from rest_framework.exceptions import ErrorDetail, ValidationError

from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    DebusineTaskConfiguration,
)
from debusine.assets import AssetCategory, asset_data_model
from debusine.client.models import (
    Link,
    LookupChildType,
    LookupResultType,
    WorkspaceInheritanceChainElement,
)
from debusine.db.context import context
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    Asset,
    Collection,
    CollectionItem,
    CollectionRelation,
    CollectionTokenGrant,
    File,
    Group,
    Token,
    User,
    WorkRequest,
    WorkflowTemplate,
    Workspace,
)
from debusine.db.models.auth import GroupQuerySet
from debusine.db.models.collections import validate_collection_name
from debusine.db.models.lookups import LookupResult
from debusine.server.exceptions import (
    DebusineAPIException,
    raise_workspace_not_found,
)
from debusine.server.workflows import Workflow
from debusine.server.workflows.models import WorkRequestManualUnblockAction
from debusine.tasks.models import EventReactions, OutputData, WorkerType

if TYPE_CHECKING:
    BaseSerializerBase: TypeAlias = serializers.BaseSerializer
    SerializerMixinBase: TypeAlias = serializers.Serializer
    RelatedFieldBase: TypeAlias = serializers.RelatedField
    SlugRelatedFieldBase: TypeAlias = serializers.SlugRelatedField
    FieldBase: TypeAlias = serializers.Field
else:
    # REST framework's serializers don't support generic types at run-time
    # yet.
    from debusine.utils.typing_utils import patch_runtime_generic_support

    BaseSerializerBase = patch_runtime_generic_support(
        serializers.BaseSerializer
    )
    SerializerMixinBase = patch_runtime_generic_support(serializers.Serializer)
    RelatedFieldBase = patch_runtime_generic_support(serializers.RelatedField)
    SlugRelatedFieldBase = patch_runtime_generic_support(
        serializers.SlugRelatedField
    )
    FieldBase = patch_runtime_generic_support(serializers.Field)


class DebusineSerializerMixin[T](SerializerMixinBase[T]):
    """Adjust a serializer with debusine-specific behaviour."""

    validation_error_title: str

    @override
    def is_valid(self, *, raise_exception: bool = False) -> bool:
        """Raise a suitable API exception when deserialization fails."""
        valid = super().is_valid()
        if raise_exception and not valid:
            if (workspace_errors := self.errors.get("workspace", [])) and all(
                isinstance(error, ErrorDetail)
                and error.code == "does_not_exist"
                for error in workspace_errors
            ):
                raise_workspace_not_found(self.data["workspace"])
            else:
                raise DebusineAPIException(
                    title=self.validation_error_title,
                    validation_errors=self.errors,
                )
        return valid


class WorkspaceInCurrentScopeField(SlugRelatedFieldBase[Workspace]):
    """A field for workspaces in the current scope."""

    def __init__(self, **kwargs: Any) -> None:
        """Construct the field."""
        kwargs.setdefault("default", self._get_default_workspace)
        super().__init__(**kwargs)

    @override
    def get_queryset(self) -> QuerySet[Workspace]:
        """Limit to the current scope."""
        return Workspace.objects.in_current_scope()

    def _get_default_workspace(self) -> Workspace:
        """Get the default workspace."""
        try:
            return self.get_queryset().get(
                name=settings.DEBUSINE_DEFAULT_WORKSPACE
            )
        except Workspace.DoesNotExist:
            # The default workspace name might not exist in the current
            # scope.  If so, the client must provide a workspace.
            self.fail("required")


class LinksSerializerMixin[T](SerializerMixinBase[T]):
    """Add a links record to a serializer."""

    def get_links(self, instance: T) -> dict[str, Link]:  # noqa: ARG002, U100
        """Return links for this object."""
        return {}

    @override
    def to_representation(self, instance: T) -> dict[str, Any]:
        """Add links to the serialized representation."""
        links: dict[str, str | dict[str, str]] = {}
        for name, link in self.get_links(instance).items():
            if link.can_be_str():
                links[name] = link.href
            else:
                links[name] = link.model_dump(exclude_unset=True)
        res = super().to_representation(instance)
        res["links"] = links
        return res


class WorkRequestSerializer(
    LinksSerializerMixin[WorkRequest],
    DebusineSerializerMixin[WorkRequest],
    serializers.ModelSerializer[WorkRequest],
):
    """Serializer of a WorkRequest."""

    validation_error_title = "Cannot deserialize work request"

    url = serializers.URLField(source="get_absolute_url", read_only=True)
    artifacts = serializers.SerializerMethodField("artifacts_for_work_request")
    scope = serializers.CharField(source="workspace.scope.name", read_only=True)
    workspace = WorkspaceInCurrentScopeField(slug_field="name")
    workflow_data = serializers.JSONField(
        source="workflow_data_json", required=False
    )
    event_reactions = serializers.JSONField(
        source='event_reactions_json', required=False
    )
    scheduler_tags_provided = serializers.ListField(
        child=serializers.CharField(), read_only=True
    )
    scheduler_tags_required = serializers.ListField(
        child=serializers.CharField(), read_only=True
    )

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize object.

        :param kwargs['only_fields']: if not None: any field in .data that does
          not will make validate() raise a ValidationError.
        """
        self._only_fields = kwargs.pop('only_fields', None)
        super().__init__(*args, **kwargs)

    @override
    def get_links(self, instance: WorkRequest) -> dict[str, Link]:
        """Get links to add to the serialized record."""
        res = super().get_links(instance)
        for link, work_request in (
            ("self", instance),
            ("parent", instance.parent),
            ("workflow_root", instance.workflow_root),
        ):
            if work_request is not None:
                res[link] = Link(
                    href=reverse(
                        "api:work-request-detail",
                        kwargs={"work_request_id": work_request.pk},
                    )
                )
                res[f"webui_{link}"] = Link(
                    href=work_request.get_absolute_url()
                )
        for link, user in (
            ("created_by", instance.created_by),
            ("aborted_by", instance.aborted_by),
        ):
            if user is not None:
                res[f"webui_{link}"] = Link(
                    href=user.get_absolute_user_id_url()
                )
        return res

    def artifacts_for_work_request(
        self, work_request: WorkRequest
    ) -> list[int]:
        """Return list of artifact ids created by this work request."""
        return list(
            work_request.artifact_set.order_by("id").values_list(
                "id", flat=True
            )
        )

    @override
    def validate(self, data: Any) -> Any:
        """
        Add extra validation in the WorkRequestSerializer. Used by .is_valid().

        -if only_fields was passed in the __init__: raise ValidationError
          if the initial data had an unexpected field.
        """
        if self._only_fields is not None:
            fields_in_data = set(self.initial_data.keys())
            wanted_fields = set(self._only_fields)

            if unwanted_fields := (fields_in_data - wanted_fields):
                raise serializers.ValidationError(
                    f"Invalid fields: {', '.join(sorted(unwanted_fields))}"
                )

        return data

    def validate_event_reactions(self, value: Any) -> Any:
        """Validate event reactions."""
        try:
            EventReactions.model_validate(value)
        except pydantic.ValidationError as e:
            raise serializers.ValidationError(f"Invalid event_reactions: {e}")
        return value

    class Meta:
        model = WorkRequest
        fields = [
            'id',
            'url',
            'parent',
            'workflow_root',
            'task_name',
            'created_at',
            'created_by',
            'completed_by',
            'aborted_by',
            'started_at',
            'completed_at',
            'duration',
            'worker',
            'task_type',
            'task_data',
            'dynamic_task_data',
            'priority_base',
            'priority_adjustment',
            'workflow_data',
            'event_reactions',
            'status',
            'result',
            'artifacts',
            'scope',
            'workspace',
            'scheduler_tags_provided',
            'scheduler_tags_required',
        ]


class WorkRequestSerializerWithConfiguredTaskData(WorkRequestSerializer):
    """Specialization of WorkRequestSerializer to feed task data to workers."""

    @override
    def to_representation(self, instance: WorkRequest) -> dict[str, Any]:
        """
        Use configured_task_data instead of task_data.

        This is intended to feed configured_task_data to workers without
        breaking the worker API.
        """
        res = super().to_representation(instance)
        if instance.configured_task_data:
            res["task_data"] = instance.configured_task_data
        return res


class WorkerWorkRequestSerializer(WorkRequestSerializerWithConfiguredTaskData):
    """Specialization of WorkRequestSerializer including a limited token."""

    token = serializers.SerializerMethodField()

    def get_token(self, instance: WorkRequest) -> str:
        """Issue a limited token and return its key."""
        return Token.objects.create_work_request_token(instance).key

    class Meta:
        model = WorkRequest
        fields = [
            *WorkRequestSerializerWithConfiguredTaskData.Meta.fields,
            "token",
        ]
        read_only_fields = fields


class WorkRequestUpdateSerializer(WorkRequestSerializer):
    """Serializer for updating work requests."""

    validation_error_title = "Cannot deserialize work request update"

    class Meta:
        model = WorkRequest
        fields = ['priority_adjustment']

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Ensure that only expected fields are passed."""
        kwargs["only_fields"] = self.Meta.fields
        super().__init__(*args, **kwargs)


class WorkRequestUnblockSerializer(
    DebusineSerializerMixin[WorkRequest], serializers.Serializer[WorkRequest]
):
    """Serializer for the data used to edit a manual unblock."""

    validation_error_title = "Cannot deserialize work request unblock"

    notes = serializers.CharField(max_length=65536, required=False)
    action = serializers.ChoiceField(
        choices=list(WorkRequestManualUnblockAction), required=False
    )

    @override
    def validate(self, data: Any) -> Any:
        """Ensure at least one of notes and action is present."""
        if not data.get("notes") and data.get("action") is None:
            raise serializers.ValidationError(
                "At least one of notes and action must be set"
            )

        return data


class WorkRequestConfirmSerializer(
    DebusineSerializerMixin[WorkRequest], serializers.Serializer[WorkRequest]
):
    """Serializer for the data used to confirm a work request."""

    validation_error_title = "Cannot deserialize work request confirmation"

    confirm = serializers.BooleanField(required=False, default=True)
    comment = serializers.CharField(required=False, max_length=65536)


class WorkRequestExternalDebsignSerializer(
    WorkRequestSerializerWithConfiguredTaskData
):
    """
    Serializer for an `ExternalDebsign` work request.

    This is similar to normal work request serialization, except that we
    make sure to return up-to-date dynamic task data so that it can be used
    to get the unsigned artifact ID.
    """

    @override
    def to_representation(self, instance: WorkRequest) -> dict[str, Any]:
        """Augment the representation with current dynamic task data."""
        data = super().to_representation(instance)

        try:
            task = instance.get_task()
            dynamic_task_data = task.compute_dynamic_data()
            data["dynamic_task_data"] = (
                None
                if dynamic_task_data is None
                else dynamic_task_data.model_dump(
                    mode="json", exclude_unset=True
                )
            )
        except Exception as exc:
            raise ValidationError(detail=f"Cannot process task data: {exc}")

        return data


class WorkRequestExternalDebsignRequestSerializer(
    DebusineSerializerMixin[Never], serializers.Serializer[Never]
):
    """Serializer for the data used to complete an `ExternalDebsign` task."""

    validation_error_title = "Cannot deserialize work request"

    signed_artifact = serializers.PrimaryKeyRelatedField(
        queryset=Artifact.objects
    )

    def validate_signed_artifact(self, value: Any) -> Any:
        """Check that the signed artifact has the correct category."""
        if value.category != ArtifactCategory.UPLOAD:
            raise ValidationError(
                detail=(
                    f"Expected signed artifact of category "
                    f"{ArtifactCategory.UPLOAD}; got {value.category}"
                )
            )
        if value.created_by_work_request is not None:
            raise ValidationError(
                detail=(
                    "Signed artifact must not have been created by a work "
                    "request"
                )
            )
        return value


class WorkerRegisterSerializer(serializers.Serializer[Never]):
    """Serializer for the data when a worker is registering."""

    token = serializers.CharField(max_length=64)
    fqdn = serializers.CharField(max_length=400)
    worker_type = serializers.ChoiceField(
        choices=list(WorkerType), default=WorkerType.EXTERNAL
    )


class WorkRequestCompletedSerializer(serializers.ModelSerializer[WorkRequest]):
    """Serializer for the data when a work request is completed."""

    result = serializers.ChoiceField(
        choices=[
            choice
            for choice in WorkRequest.Results
            if choice != WorkRequest.Results.NONE
        ]
    )
    output_data = serializers.JSONField(required=False)

    def validate_output_data(self, value: Any) -> Any:
        """Validate output_data."""
        try:
            OutputData.model_validate(value)
        except pydantic.ValidationError as e:
            raise serializers.ValidationError(f"Invalid output_data: {e}")
        return value

    class Meta:
        model = WorkRequest
        fields = ["result", "output_data"]


class OnWorkRequestCompleted(serializers.Serializer[Never]):
    """Serializer used by the OnWorkRequestCompletedConsumer."""

    type = serializers.CharField()
    text = serializers.CharField()
    work_request_id = serializers.IntegerField()
    completed_at = serializers.DateTimeField()
    result = serializers.ChoiceField(choices=list(WorkRequest.Results))


class FileSerializer(serializers.Serializer[File]):
    """Serializer for the File class."""

    size = serializers.IntegerField(min_value=0)
    checksums = serializers.DictField(
        child=serializers.CharField(max_length=255)
    )
    type = serializers.ChoiceField(choices=("file",))
    content_type = serializers.CharField(allow_null=True, required=False)


class FileSerializerResponse(FileSerializer):
    """Serializer for the File responses. Include URL."""

    url = serializers.URLField()


class ArtifactSerializer(
    DebusineSerializerMixin[Artifact], serializers.Serializer[Artifact]
):
    """Serializer to deserialize the client request to create an artifact."""

    validation_error_title = "Cannot deserialize artifact"

    category = serializers.CharField(max_length=255)
    workspace = WorkspaceInCurrentScopeField(slug_field="name")
    files = serializers.DictField(child=FileSerializer())
    # Deliberately incompatible with Serializer.data.
    data = serializers.DictField()  # type: ignore[assignment]
    work_request = serializers.IntegerField(allow_null=True, required=False)
    expiration_delay = serializers.DurationField(
        allow_null=True, required=False
    )


class ArtifactSerializerResponse(
    LinksSerializerMixin[Artifact], serializers.ModelSerializer[Artifact]
):
    """Serializer for returning the information of an Artifact."""

    url = serializers.URLField(source="get_absolute_url", read_only=True)

    # Artifact's scope and workspace
    scope = serializers.CharField(source="workspace.scope.name", read_only=True)
    workspace = serializers.CharField(source="workspace.name", read_only=True)

    # Files with their hash and size
    files = serializers.SerializerMethodField("_files")

    # List of file paths pending to be uploaded
    files_to_upload = serializers.SerializerMethodField("_files_to_upload")

    # URL to download the artifact
    download_tar_gz_url = serializers.SerializerMethodField(
        "_download_tar_gz_url"
    )

    _base_download_url: str

    class Meta:
        model = Artifact
        fields = [
            "id",
            "url",
            "scope",
            "workspace",
            "category",
            "data",
            "created_at",
            "expire_at",
            "download_tar_gz_url",
            "files_to_upload",
            "files",
        ]

    @override
    def get_links(self, instance: Artifact) -> dict[str, Link]:
        """Get links to add to the serialized record."""
        res = super().get_links(instance)
        res["self"] = Link(
            href=reverse(
                "api:artifact",
                kwargs={"artifact_id": instance.pk},
            )
        )
        res["webui_self"] = Link(href=instance.get_absolute_url())
        res["download_tar_gz"] = Link(
            href=instance.get_absolute_url_download() + "?archive=tar.gz",
            type="application/tar+gzip",
        )
        return res

    def _files_to_upload(self, artifact: Artifact) -> list[str]:
        """Return file paths from artifact that have not been uploaded yet."""
        return list(
            artifact.fileinartifact_set.filter(complete=False)
            .order_by("path")
            .values_list("path", flat=True)
        )

    def _files(self, artifact: Artifact) -> dict[str, dict[str, Any]]:
        """
        Return files in the artifact with their hash and size.

        The files might not be available yet if they have not been uploaded.
        """
        files: dict[str, dict[str, Any]] = {}

        for file_in_artifact in artifact.fileinartifact_set.order_by("path"):
            file = file_in_artifact.file
            data = {
                "size": file.size,
                "type": "file",
                "checksums": {
                    file.current_hash_algorithm: file.hash_digest.hex()
                },
                "url": self._base_download_url + file_in_artifact.path,
                "content_type": file_in_artifact.content_type,
            }
            file_serialized = FileSerializerResponse(data=data)
            file_serialized.is_valid(raise_exception=True)

            files[file_in_artifact.path] = file_serialized.data

        return files

    def _download_tar_gz_url(self, obj: Artifact) -> str:  # noqa: ARG002, U100
        """Based on self._base_download_url return URL to download .tar.gz."""
        return self._base_download_url + "?archive=tar.gz"

    @staticmethod
    def _build_absolute_download_url(
        artifact: Artifact, request: HttpRequest
    ) -> str:
        """Return URL to download the artifact (without the archive format)."""
        return request.build_absolute_uri(artifact.get_absolute_url_download())

    @classmethod
    def from_artifact(
        cls, artifact: Artifact, request: HttpRequest
    ) -> "ArtifactSerializerResponse":
        """Given an artifact and request return ArtifactSerializerResponse."""
        serialized = ArtifactSerializerResponse(artifact)
        serialized._base_download_url = cls._build_absolute_download_url(
            artifact, request
        )
        return serialized


class ArtifactRelationResponseSerializer(
    DebusineSerializerMixin[ArtifactRelation],
    serializers.ModelSerializer[ArtifactRelation],
):
    """Serializer for relations between artifacts."""

    validation_error_title = "Cannot deserialize artifact relation"

    media_type = "application/json"

    class Meta:
        model = ArtifactRelation
        fields = ["id", "artifact", "target", "type"]
        # Don't let REST framework validate uniqueness constraints (which it
        # does automatically as of 3.15.0).  We want GetOrCreateAPIView to
        # return 200 if the relation already exists, which it will do if
        # REST framework hasn't returned 400 first due to a validation
        # failure.
        validators: list[Callable[[ArtifactRelation], None]] = []


class AssetSerializer(
    DebusineSerializerMixin[Asset], serializers.Serializer[Asset]
):
    """Serializer to deserialize the client request to create an asset."""

    validation_error_title = "Cannot deserialize asset"

    id = serializers.IntegerField(required=False, read_only=True)
    created_at = serializers.DateTimeField(read_only=True)
    category = serializers.CharField(max_length=255)
    scope = serializers.CharField(source="workspace.scope.name", read_only=True)
    workspace = WorkspaceInCurrentScopeField(slug_field="name")
    # Deliberately incompatible with Serializer.data.
    data = serializers.DictField()  # type: ignore[assignment]
    work_request = serializers.IntegerField(
        allow_null=True, required=False, source="created_by_work_request_id"
    )

    @override
    def validate(self, data: dict[str, Any]) -> dict[str, Any]:
        """Validate category and data."""
        if data["category"] not in AssetCategory:
            raise serializers.ValidationError(
                {"category": f"{data['category']} is not a known category."}
            )

        try:
            asset_data_model(data["category"], data["data"])
        except ValueError as e:
            raise serializers.ValidationError(
                {"data": f"invalid asset data: {e}"}
            ) from e
        return data

    @override
    def create(self, validated_data: dict[str, Any]) -> Asset:
        """Create an Asset model from deserialized data."""
        data = copy(validated_data)
        if not data.get("created_by_work_request_id"):
            data["created_by"] = context.user
        return Asset.objects.create(**data)


class AssetPermissionCheckRequest(
    DebusineSerializerMixin[Never], serializers.Serializer[Never]
):
    """Serializer for a permission check on an asset."""

    validation_error_title = "Cannot deserialize permission check request"

    artifact_id = serializers.IntegerField()
    work_request_id = serializers.IntegerField()
    workspace = serializers.CharField()


class AssetPermissionCheckResponse(
    DebusineSerializerMixin[Never], serializers.Serializer[Never]
):
    """Serializer for a permission check result."""

    validation_error_title = "Cannot serialize permission check result"

    has_permission = serializers.BooleanField()
    username = serializers.CharField(max_length=150, required=False)
    user_id = serializers.IntegerField(required=False)
    resource = serializers.DictField(
        child=serializers.CharField(max_length=255), required=False
    )


class LookupExpectTypeChoiceField(serializers.ChoiceField):
    """
    A field for `Lookup*.expect_type`.

    This handles values of `LookupChildType`, but also accepts values of
    `CollectionItem.Types` for compatibility with old clients.
    """

    def __init__(self, **kwargs: Any) -> None:
        """Initialize the field."""
        super().__init__(choices=list(LookupChildType))

    @override
    def to_internal_value(self, data: Any) -> str:
        """Accept values of `CollectionItemTypes` for compatibility."""
        if data in CollectionItem.Types:
            data = {
                CollectionItem.Types.BARE: LookupChildType.BARE,
                CollectionItem.Types.ARTIFACT: LookupChildType.ARTIFACT,
                CollectionItem.Types.COLLECTION: LookupChildType.COLLECTION,
            }[data]
        return super().to_internal_value(data)


class LookupSingleSerializer(serializers.Serializer[Never]):
    """Serializer for the data used to look up a single collection item."""

    lookup = serializers.JSONField()
    default_category = serializers.CharField(
        max_length=255, required=False, allow_null=True
    )
    workspace = WorkspaceInCurrentScopeField(slug_field="name")
    workflow_root = serializers.PrimaryKeyRelatedField(
        queryset=WorkRequest.objects, required=False, allow_null=True
    )
    expect_type = LookupExpectTypeChoiceField()


class LookupMultipleSerializer(serializers.Serializer[Never]):
    """Serializer for the data used to look up multiple collection items."""

    lookup = serializers.JSONField()
    default_category = serializers.CharField(
        max_length=255, required=False, allow_null=True
    )
    workspace = WorkspaceInCurrentScopeField(slug_field="name")
    workflow_root = serializers.PrimaryKeyRelatedField(
        queryset=WorkRequest.objects, required=False, allow_null=True
    )
    expect_type = LookupExpectTypeChoiceField()


class CollectionItemSerializer(serializers.Serializer[CollectionItem]):
    """Serialize the collection item for nesting in the response."""

    name = serializers.CharField()
    # Deliberately incompatible with Serializer.data.
    data = serializers.JSONField(allow_null=True)  # type: ignore[assignment]


class LookupResponseSerializer(
    LinksSerializerMixin[LookupResult], serializers.Serializer[LookupResult]
):
    """Serializer for the response to looking up collection items."""

    media_type = "application/json"

    result_type = serializers.ChoiceField(choices=list(LookupResultType))
    collection_item = CollectionItemSerializer(read_only=True, allow_null=True)
    parent_collection_lookup = serializers.CharField(
        read_only=True, allow_null=True
    )
    artifact: serializers.PrimaryKeyRelatedField[Artifact] = (
        serializers.PrimaryKeyRelatedField(read_only=True, allow_null=True)
    )
    collection: serializers.PrimaryKeyRelatedField[Collection] = (
        serializers.PrimaryKeyRelatedField(read_only=True, allow_null=True)
    )

    @override
    def get_links(self, instance: LookupResult) -> dict[str, Link]:
        """Get links to add to the serialized record."""
        res = super().get_links(instance)
        result: Artifact | Collection | None = None
        self_link: Link | None = None
        match instance.result_type:
            case LookupResultType.ARTIFACT:
                result = instance.artifact
                assert result is not None
                self_link = Link(
                    href=reverse(
                        "api:artifact", kwargs={"artifact_id": result.pk}
                    )
                )
            case LookupResultType.COLLECTION:
                result = instance.collection
                assert result is not None
                self_link = Link(
                    href=reverse(
                        "api:collection-detail",
                        kwargs={
                            "workspace": result.workspace.name,
                            "pk": result.pk,
                        },
                    )
                )
        if result is not None:
            assert self_link is not None
            res["self"] = self_link
            res["webui_self"] = Link(href=result.get_absolute_url())
        if instance.parent_collection_lookup:
            assert instance.collection_item is not None
            parent_collection = instance.collection_item.parent_collection
            assert parent_collection is not None
            res["parent_collection"] = Link(
                href=reverse(
                    "api:collection-detail",
                    kwargs={
                        "workspace": parent_collection.workspace.name,
                        "pk": parent_collection.pk,
                    },
                )
            )
            res["webui_parent_collection"] = Link(
                href=parent_collection.get_absolute_url()
            )
        return res


class ExistingGroupSerializer(RelatedFieldBase[Group, int | str, str]):
    """Read-only group reference, by ID or name."""

    default_error_messages = {
        "does_not_exist": "Group {value} does not exist.",
        "invalid": "Invalid value.",
    }

    @override
    def to_internal_value(self, data: int | str) -> Group:
        queryset = self.get_queryset()
        assert isinstance(queryset, GroupQuerySet)
        try:
            match data:
                case int():
                    res = queryset.get(pk=data)
                    assert isinstance(res, Group)
                    return res
                case str():
                    return queryset.get_by_name(data)
                case _:
                    self.fail('invalid')
        except ObjectDoesNotExist:
            self.fail('does_not_exist', value=smart_str(data))
        except (TypeError, ValueError):
            self.fail('invalid')

    # obj should be Group | PKOnlyObject according to rest framework stubs, but
    # I'm not sure how to express that. Using a type: ignore for now
    @override
    def to_representation(self, obj: Group) -> str:  # type: ignore[override]
        return str(obj)


class WorkflowTemplateSerializer(
    LinksSerializerMixin[WorkflowTemplate],
    DebusineSerializerMixin[WorkflowTemplate],
    serializers.ModelSerializer[WorkflowTemplate],
):
    """Serializer for workflow templates."""

    validation_error_title = "Cannot deserialize workflow template"

    media_type = "application/json"

    extra_groups = ExistingGroupSerializer(
        many=True, required=False, queryset=Group.objects.all()
    )

    class Meta:
        model = WorkflowTemplate
        fields = [
            "id",
            "name",
            "task_name",
            "static_parameters",
            "runtime_parameters",
            "priority",
            "extra_groups",
            "restricted",
        ]

    @override
    def to_representation(self, instance: WorkflowTemplate) -> dict[str, Any]:
        """Sort fields according to the model order."""
        res = super().to_representation(instance)

        try:
            workflow_cls = Workflow.from_name(instance.task_name)
            expected_order = list(
                workflow_cls.task_data_type.model_fields.keys()
            )
        except (ValueError, AttributeError):
            return res

        for field_name in ["static_parameters", "runtime_parameters"]:
            if field_name in res and isinstance(res[field_name], dict):
                params = res[field_name]

                sorted_params = {
                    key: params[key] for key in expected_order if key in params
                }
                for key, value in params.items():
                    if key not in sorted_params:
                        sorted_params[key] = value
                res[field_name] = sorted_params

        res["extra_groups"].sort()

        return res

    @override
    def get_links(self, instance: WorkflowTemplate) -> dict[str, Link]:
        """Get links to add to the serialized record."""
        res = super().get_links(instance)
        res["self"] = Link(
            href=reverse(
                "api:workflow-template-detail",
                kwargs={
                    "workspace": instance.workspace.name,
                    "pk": instance.pk,
                },
            )
        )
        res["webui_self"] = Link(href=instance.get_absolute_url())
        return res

    def validate_task_name(self, value: str) -> str:
        """
        Ensure that task_name is valid.

        We don't currently validate task_data, since a template's task data
        is allowed to be incomplete.
        """
        try:
            Workflow.from_name(value)
        except ValueError as exc:
            raise serializers.ValidationError({"task_name": str(exc)})

        return value


class CreateWorkflowRequestSerializer(
    DebusineSerializerMixin[WorkRequest],
    serializers.ModelSerializer[WorkRequest],
):
    """Serializer for requests to create workflows."""

    validation_error_title = "Cannot deserialize workflow creation request"

    media_type = "application/json"

    template_name = serializers.CharField(
        default="", max_length=255, allow_blank=True
    )
    task_name = serializers.CharField(
        default="", max_length=255, allow_blank=True
    )
    workspace = WorkspaceInCurrentScopeField(slug_field="name")

    class Meta:
        model = WorkRequest
        fields = [
            "template_name",
            "task_name",
            "workspace",
            "task_data",
        ]

    @override
    def validate(self, data: Any) -> Any:
        """Exactly one of ``template_name`` and ``task_name`` must be set."""
        if bool(data.get("template_name")) == bool(data.get("task_name")):
            raise serializers.ValidationError(
                "Exactly one of template_name and task_name must be set"
            )
        return data


class PydanticSerializerMixin[PydanticModel: pydantic.BaseModel](
    SerializerMixinBase[PydanticModel]
):
    """Serialize pydantic models."""

    pydantic_model: type[PydanticModel]

    @override
    def to_representation(self, instance: PydanticModel) -> dict[str, Any]:
        """Delegate representation to pydantic."""
        return instance.model_dump()

    @override
    def to_internal_value(self, data: Any) -> PydanticModel:
        """Delegate validation to pydantic."""
        return self.pydantic_model(**data)


class DebusineTaskConfigurationSerializer(
    DebusineSerializerMixin[DebusineTaskConfiguration],
    PydanticSerializerMixin[DebusineTaskConfiguration],
):
    """Serialize a DebusineTaskConfiguration item."""

    pydantic_model = DebusineTaskConfiguration


class TaskConfigurationCollectionSerializer(
    DebusineSerializerMixin[Collection], serializers.ModelSerializer[Collection]
):
    """Serializer for a TaskConfiguration collection."""

    validation_error_title = "Cannot deserialize task configuration collection"
    media_type = "application/json"

    id = serializers.IntegerField()

    class Meta:
        model = Collection
        fields = ["id", "name", "data"]


class TaskConfigurationCollectionContents(NamedTuple):
    """Bundle together a TaskConfiguration collection and its items."""

    collection: Collection
    items: list[DebusineTaskConfiguration]


class TaskConfigurationCollectionContentsSerializer(
    DebusineSerializerMixin[TaskConfigurationCollectionContents],
):
    """Serializer for the contents of a TaskConfiguration collection."""

    validation_error_title = "Cannot deserialize task configuration collection"
    media_type = "application/json"

    collection = TaskConfigurationCollectionSerializer()
    items = DebusineTaskConfigurationSerializer(many=True)

    def validate_items(
        self, value: list[DebusineTaskConfiguration]
    ) -> list[DebusineTaskConfiguration]:
        """Ensure the collection items do not have conflicting names."""
        names: set[str] = set()
        for item in value:
            if (name := item.name()) in names:
                raise serializers.ValidationError(
                    f"{name!r}: found multiple items with the same name"
                )
            names.add(name)
        return value


class TaskConfigurationCollectionUpdateResultsSerializer(
    DebusineSerializerMixin[dict[str, int]],
    serializers.Serializer[dict[str, int]],
):
    """Serializer for the result information of a task configuration update."""

    validation_error_title = (
        "Cannot deserialize task configuration collection update results"
    )

    added = serializers.IntegerField()
    updated = serializers.IntegerField()
    removed = serializers.IntegerField()
    unchanged = serializers.IntegerField()


class WorkspaceChainItemField(
    FieldBase[Workspace, dict[str, Any], dict[str, Any], Workspace]
):
    """Serializers for a Workspace in a workspace chain."""

    @override
    def to_representation(self, value: Workspace) -> dict[str, Any]:
        """Trivially serialize workspaces with IDs."""
        return WorkspaceInheritanceChainElement(
            id=value.id,
            scope=value.scope.name,
            workspace=value.name,
        ).model_dump(mode="json")

    def _workspace_from_id(
        self, el: WorkspaceInheritanceChainElement
    ) -> Workspace:
        assert el.id is not None

        workspace_qs = Workspace.objects.can_display(context.pc).select_related(
            "scope"
        )
        try:
            workspace = workspace_qs.get(pk=el.id)
        except Workspace.DoesNotExist:
            raise ValidationError(f"Workspace with id {el.id} does not exist")
        if el.workspace is not None and workspace.name != el.workspace:
            raise ValidationError(
                f"Workspace with id {el.id} is not named {el.workspace!r}"
            )
        if el.scope is not None and workspace.scope.name != el.scope:
            raise ValidationError(
                f"Workspace with id {el.id} is not in scope {el.scope!r}"
            )
        return workspace

    def _workspace_from_scope_workspace(
        self, el: WorkspaceInheritanceChainElement
    ) -> Workspace:
        assert el.workspace is not None

        workspace_qs = Workspace.objects.can_display(context.pc).select_related(
            "scope"
        )

        if el.scope is None:
            el.scope = context.require_scope().name

        try:
            return workspace_qs.get(scope__name=el.scope, name=el.workspace)
        except Workspace.DoesNotExist:
            raise ValidationError(
                f"Workspace {el.workspace!r}"
                f" does not exist in scope {el.scope!r}"
            )

    @override
    def to_internal_value(self, data: Any) -> Workspace:
        """Parse a workspace from IDs or ``scope/name`` strings."""
        if not isinstance(data, dict):
            raise ValidationError("Workspace indicator should be a dict")

        try:
            parsed = WorkspaceInheritanceChainElement(**data)
        except ValueError as e:
            raise ValidationError(str(e))

        if parsed.id is not None:
            return self._workspace_from_id(parsed)
        else:
            return self._workspace_from_scope_workspace(parsed)


class WorkspaceChainSerializer(serializers.Serializer[Never]):
    """Serializer for a workspace chain."""

    chain = serializers.ListField(
        # Set an arbitrary high value to prevent a misguided API call impact
        # server performance
        child=WorkspaceChainItemField(),
        max_length=100,
    )


class DurationInDays(FieldBase[dt.timedelta, int, int, dt.timedelta]):
    """Represent a duration field reduced to a number of days."""

    @override
    def to_representation(self, value: dt.timedelta) -> int:
        """Represent a timedelta as a number of days."""
        days = value.days
        if value.seconds > 0 or value.microseconds > 0:
            days += 1
        return days

    @override
    def to_internal_value(self, data: int) -> dt.timedelta:
        """Parse a timedelta from a number of days."""
        return dt.timedelta(days=int(data))


class CollectionSerializer(
    LinksSerializerMixin[Collection],
    DebusineSerializerMixin[Collection],
    serializers.ModelSerializer[Collection],
):
    """Serializer of a Collection."""

    validation_error_title = "Cannot deserialize collection"

    category = serializers.ChoiceField(
        choices=list(CollectionCategory), required=False
    )
    name = serializers.CharField(required=False)
    full_history_retention_period = DurationInDays(
        allow_null=True, required=False
    )
    metadata_only_retention_period = DurationInDays(
        allow_null=True, required=False
    )
    stats = serializers.SerializerMethodField()

    @override
    def get_links(self, instance: Collection) -> dict[str, Link]:
        """Get links to add to the serialized record."""
        res = super().get_links(instance)
        res["self"] = Link(
            href=reverse(
                "api:collection-detail",
                kwargs={
                    "workspace": instance.workspace.name,
                    "pk": instance.pk,
                },
            )
        )
        res["webui_self"] = Link(href=instance.get_absolute_url())
        res["webui_category"] = Link(
            href=reverse(
                "workspaces:collections:category_list",
                kwargs={
                    "wname": instance.workspace.name,
                    "ccat": instance.category,
                },
            ),
            type="category",
        )
        return res

    def get_stats(self, instance: Collection) -> list[dict[str, Any]] | None:
        """Get statistics on the collections if requested."""
        if not self.context.get("with_stats"):
            return None
        return [
            {
                "type": x["type"].name.lower(),
                "category": x["category"],
                "active": x["active"],
                "count": x["count"],
            }
            for x in instance.get_item_stats()
        ]

    def validate_name(self, value: str) -> str:
        """Ensure that name is valid."""
        validate_collection_name(value)
        return value

    class Meta:
        model = Collection
        fields = [
            "id",
            "category",
            "name",
            "full_history_retention_period",
            "metadata_only_retention_period",
            "data",
            "stats",
        ]


class CollectionIdentifierSerializer(serializers.Serializer[Any]):
    """Identify collections based on the category and name."""

    name = serializers.CharField()
    category = serializers.ChoiceField(choices=list(CollectionCategory))


class CollectionRelationSerializer(
    DebusineSerializerMixin[CollectionRelation],
    serializers.ModelSerializer[CollectionRelation],
):
    """Serializer for relations between collections."""

    validation_error_title = "Cannot deserialize collection relation"

    # The attribute name "source" cannot be used (mypy complains). Use
    # "source_collection", use "source" parameter to map to the API field
    source_collection = CollectionIdentifierSerializer(source="source")
    target_collection = CollectionIdentifierSerializer(source="target")

    class Meta:
        model = CollectionRelation
        fields = [
            "source_collection",
            "target_collection",
            "type",
            "position",
        ]


class CollectionRelationSetSerializer(serializers.Serializer[Any]):
    """
    Serializer for setting the targets for a relation type.

    Used for the request and response.
    """

    targets = CollectionIdentifierSerializer(many=True)


class CollectionTokenGrantSerializer(
    serializers.ModelSerializer[CollectionTokenGrant]
):
    """Serializer for a CollectionTokenGrant."""

    token_type = serializers.CharField(
        source="token.token_type", required=False
    )
    token_hash = serializers.CharField(
        source="token.hash", default="", max_length=64, allow_blank=True
    )
    token_key = serializers.CharField(source="token.key", required=False)
    username = serializers.CharField(
        source="token.get_username",
        default="",
        max_length=150,
        allow_blank=True,
    )
    comment = serializers.CharField(
        source="token.comment", default="", max_length=100, allow_blank=True
    )

    class Meta:
        model = CollectionTokenGrant
        fields = [
            "token_type",
            "token_hash",
            "token_key",
            "username",
            "comment",
            "role",
        ]

    @override
    def validate(self, data: dict[str, Any]) -> dict[str, Any]:
        """Validate consistency of token-related fields."""
        if not data["token"]["hash"] and not data["token"]["get_username"]:
            raise ValidationError("Must provide either token_hash or username")
        return data

    @override
    def create(self, validated_data: dict[str, Any]) -> CollectionTokenGrant:
        collection = validated_data["collection"]
        role = Collection.Roles(validated_data["role"])
        created_by = validated_data["created_by"]
        token_hash = validated_data["token"]["hash"]
        username = validated_data["token"]["get_username"]
        comment = validated_data["token"]["comment"]
        assert isinstance(collection, Collection)
        assert isinstance(created_by, User)
        assert isinstance(token_hash, str)
        assert isinstance(username, str)
        assert isinstance(comment, str)

        if token_hash:
            token = Token.objects.get(hash=token_hash)
            match token.token_type:
                # It would be easy to add support for user tokens as well if
                # we wanted to.
                case Token.Types.EXTERNAL_USER:
                    if username and username != token.external_username:
                        raise ValidationError(
                            f"Token {token} has username"
                            f" {token.external_username!r}, not {username!r}"
                        )
                case _:
                    raise ValidationError(
                        f"Cannot grant roles to {token.token_type} tokens"
                    )
        else:
            token = Token.objects.create_external_user_token(
                username=username, created_by=created_by, comment=comment
            )
            token.full_clean()
            token.enable()

        return collection.grant_role(token, role, created_by=created_by)


class EmitMetricSerializer(serializers.Serializer[Never]):
    """Serializer for a metric to be emitted."""

    metric_type = serializers.ChoiceField(
        choices=("counter", "gauge", "summary", "histogram")
    )
    name = serializers.CharField()
    labels = serializers.DictField(
        child=serializers.CharField(), allow_empty=True
    )
    value = serializers.FloatField()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for permission predicates context."""

from unittest import mock

from debusine.db.models import Group
from debusine.db.permissioncontext import PermissionContext
from debusine.db.playground import scenarios
from debusine.test.django import TestCase


class PermissionContextTests(TestCase):
    """Tests for permission infrastructure."""

    scenario = scenarios.DefaultContextAPI(set_current=True)

    def test_str(self) -> None:
        external_user_token = self.playground.create_external_user_token()
        worker_activation_token = (
            self.playground.create_worker_activation_token()
        )
        worker_token = self.playground.create_worker_token()
        work_request_token = self.playground.create_work_request_token()
        assert work_request_token.work_request is not None
        pc_with_extra_groups = self.playground.create_pc_with_extra_groups(
            self.scenario.scope_owners
        )
        for value, expected in (
            (PermissionContext.for_ui_anonymous(), "AnonymousUser"),
            (
                PermissionContext.for_ui_user(self.scenario.user),
                str(self.scenario.user),
            ),
            (
                PermissionContext.for_api_user(self.scenario.user_token),
                f"{self.scenario.user}[user token]",
            ),
            (
                PermissionContext.for_api_external_user(external_user_token),
                "AnonymousUser[external user token]",
            ),
            (
                PermissionContext.for_worker(worker_token),
                "AnonymousUser[worker token]",
            ),
            (
                PermissionContext.for_worker_activation(
                    worker_activation_token
                ),
                "AnonymousUser[worker-activation token]",
            ),
            (
                PermissionContext.for_work_request(token=work_request_token),
                f"{self.scenario.user}[work-request token]"
                f"[work request {work_request_token.work_request}]",
            ),
            (
                PermissionContext.for_work_request(
                    work_request=work_request_token.work_request
                ),
                f"{self.scenario.user}"
                f"[work request {work_request_token.work_request}]",
            ),
            (
                pc_with_extra_groups,
                f"{self.scenario.user}"
                f"[work request {pc_with_extra_groups.work_request}]"
                f"[extra groups {self.scenario.scope_owners}]",
            ),
        ):
            with self.subTest(value=repr(value)):
                self.assertEqual(str(value), expected)

    def test_compare(self) -> None:
        for first, second, expected in (
            (
                PermissionContext.for_ui_anonymous(),
                PermissionContext.for_ui_anonymous(),
                True,
            ),
            (self.scenario.context_ui(), self.scenario.context_ui(), True),
            (
                self.scenario.context_ui(),
                PermissionContext.for_ui_anonymous(),
                False,
            ),
            (
                self.scenario.context_ui(),
                PermissionContext.for_api_user(self.scenario.user_token),
                False,
            ),
            (
                self.scenario.context_ui(),
                self.scenario.context_ui().with_permission_checks_disabled(),
                False,
            ),
            (
                self.scenario.context_ui(),
                42,
                False,
            ),
        ):
            with self.subTest(first=repr(first), second=repr(second)):
                self.assertEqual(first == second, expected)

    def test_groups(self) -> None:
        self.scenario.workspace_owners.users.add(self.scenario.user)
        for pc, groups in (
            (PermissionContext.for_ui_anonymous(), []),
            (self.scenario.context_ui(), [self.scenario.workspace_owners]),
            (
                self.playground.create_pc_with_extra_groups(
                    self.scenario.scope_owners
                ),
                [self.scenario.workspace_owners, self.scenario.scope_owners],
            ),
        ):
            with self.subTest(pc=str(pc)):
                self.assertQuerySetEqual(pc.groups(), groups, ordered=False)

    def test_explicit_grants_only(self) -> None:
        external_user_token = self.playground.create_external_user_token()
        worker_activation_token = (
            self.playground.create_worker_activation_token()
        )
        worker_token = self.playground.create_worker_token()
        work_request_token = self.playground.create_work_request_token()
        work_request = self.playground.create_worker_task()
        for pc in (
            PermissionContext.for_api_user(self.scenario.user_token),
            PermissionContext.for_maintenance(),
            PermissionContext.for_ui_anonymous(),
            PermissionContext.for_ui_user(self.scenario.user),
            PermissionContext.for_work_request(token=work_request_token),
            PermissionContext.for_work_request(work_request=work_request),
            PermissionContext.for_worker(worker_token),
            PermissionContext.for_worker_activation(worker_activation_token),
        ):
            with self.subTest(pc=repr(pc)):
                self.assertFalse(pc.explicit_grants_only)

        for pc in (
            PermissionContext.for_api_external_user(external_user_token),
        ):
            with self.subTest(pc=repr(pc)):
                self.assertTrue(pc.explicit_grants_only)

    def test_is_non_permission_bearing_token(self) -> None:
        external_user_token = self.playground.create_external_user_token()
        worker_activation_token = (
            self.playground.create_worker_activation_token()
        )
        worker_token = self.playground.create_worker_token()
        work_request_token = self.playground.create_work_request_token()
        work_request = self.playground.create_worker_task()
        for pc in (
            PermissionContext.for_api_external_user(external_user_token),
            PermissionContext.for_api_user(self.scenario.user_token),
            PermissionContext.for_maintenance(),
            PermissionContext.for_ui_anonymous(),
            PermissionContext.for_ui_user(self.scenario.user),
            PermissionContext.for_work_request(token=work_request_token),
            PermissionContext.for_work_request(work_request=work_request),
        ):
            with self.subTest(pc=repr(pc)):
                self.assertFalse(pc.is_non_permission_bearing_token)

        for pc in (
            PermissionContext.for_worker(worker_token),
            PermissionContext.for_worker_activation(worker_activation_token),
        ):
            with self.subTest(pc=repr(pc)):
                self.assertTrue(pc.is_non_permission_bearing_token)

    def test_for_work_request_extra_groups(self) -> None:
        group1 = self.playground.create_group("group1")
        self.enterContext(
            mock.patch(
                "debusine.db.models.WorkRequest.get_extra_groups",
                return_value=Group.objects.filter(pk=group1.pk),
            )
        )
        token = self.playground.create_work_request_token()
        assert token.work_request is not None
        self.assertQuerySetEqual(
            PermissionContext.for_work_request(token=token).extra_groups,
            (group1,),
        )
        self.assertQuerySetEqual(
            PermissionContext.for_work_request(
                work_request=token.work_request
            ).extra_groups,
            (group1,),
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""View mixins for pagination."""

from typing import Any, TYPE_CHECKING, override

from django.db.models import Model
from django.views.generic import View

from debusine.web.views.base import MultipleObjectMixinBase
from debusine.web.views.table.paginator import Paginator
from debusine.web.views.table.table import Table

if TYPE_CHECKING:
    from django.core.paginator import _SupportsPagination

    _SupportsPagination  # fake usage for vulture


class TableMixin[M: Model](MultipleObjectMixinBase[M], View):
    """Mixin for ListViews using Debusine's Paginator."""

    table_class: type[Table[M]]

    @override
    def get_paginator(
        self,
        queryset: "_SupportsPagination[M]",
        per_page: int,
        orphans: int = 0,
        allow_empty_first_page: bool = True,
        **kwargs: Any,
    ) -> Paginator[M]:
        table = self.table_class(self.request, queryset)
        return table.get_paginator(
            per_page,
            orphans=orphans,
            allow_empty_first_page=allow_empty_first_page,
            **kwargs,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""UI helpers for artifacts."""

from functools import cached_property
from typing import override

from debusine.artifacts.models import ArtifactCategory
from debusine.db.models import Artifact, FileInArtifact
from debusine.db.models.artifacts import ARTIFACT_CATEGORY_ICON_NAMES
from debusine.web.views import ui_shortcuts
from debusine.web.views.places import ResourcePlace
from debusine.web.views.ui.base import UI


class ArtifactUI(UI[Artifact]):
    """UI helpers for Artifact instances."""

    @cached_property
    def place(self) -> ResourcePlace:
        """Return a place to show this artifact."""
        return ResourcePlace(
            title=f"Artifact {self.instance.get_label()}",
            breadcrumb=self.instance.get_label(),
            url=self.instance.get_absolute_url(),
            icon=ARTIFACT_CATEGORY_ICON_NAMES.get(
                ArtifactCategory(self.instance.category), "file"
            ),
            description=f"{self.instance.category} artifact"
            f" {self.instance.get_label()}",
            parent=self.instance.workspace.ui(self.request).place,
        )

    @override
    def build_ui_shortcuts(self) -> list[ui_shortcuts.UIShortcut]:
        result = super().build_ui_shortcuts()
        result += [
            ui_shortcuts.create_artifact_view(self.instance),
            ui_shortcuts.create_artifact_download(self.instance),
        ]
        return result


class FileInArtifactUI(UI[FileInArtifact]):
    """UI helpers for FileInArtifact instances."""

    @cached_property
    def place(self) -> ResourcePlace:
        """Return a place to show this file."""
        title = str(self.instance.path)
        if not self.instance.complete:
            title += " (incomplete)"
        return ResourcePlace(
            title=title,
            breadcrumb=str(self.instance.path),
            url=self.instance.get_absolute_url(),
            parent=self.instance.artifact.ui(self.request).place,
        )

    @cached_property
    def place_raw(self) -> ResourcePlace:
        """Return a place to show this file (raw view)."""
        title = str(self.instance.path)
        if not self.instance.complete:
            title += " (incomplete)"
        return ResourcePlace(
            title=title,
            breadcrumb=str(self.instance.path),
            url=self.instance.get_absolute_url_raw(),
            parent=self.instance.artifact.ui(self.request).place,
        )

    @cached_property
    def place_download(self) -> ResourcePlace:
        """Return a place to show this file (download view)."""
        title = str(self.instance.path)
        if not self.instance.complete:
            title += " (incomplete)"
        return ResourcePlace(
            title=title,
            breadcrumb=str(self.instance.path),
            url=self.instance.get_absolute_url_download(),
            parent=self.instance.artifact.ui(self.request).place,
        )

    @override
    def build_ui_shortcuts(self) -> list[ui_shortcuts.UIShortcut]:
        result = super().build_ui_shortcuts()
        result += [
            ui_shortcuts.create_file_view(self.instance),
            ui_shortcuts.create_file_view_raw(self.instance),
            ui_shortcuts.create_file_download(self.instance),
        ]
        return result



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for workflow models."""

from abc import ABCMeta, abstractmethod
from typing import Any, override

from debusine.server.workflows.models import (
    AutopkgtestWorkflowData,
    BlhcWorkflowData,
    ChildWorkspaceData,
    DebianPipelineWorkflowData,
    LintianWorkflowData,
    PackagePublishWorkflowData,
    PiupartsWorkflowData,
    QAFailOn,
    QAWorkflowData,
    RegressionTrackingUpdateQAResults,
    RegressionTrackingWorkflowData,
    ReverseDependenciesAutopkgtestWorkflowData,
    SbuildWorkflowData,
)
from debusine.tasks.models import LookupMultiple, LookupSingle, SbuildInput
from debusine.test.django import TestCase


class SbuildWorkflowDataTests(TestCase):
    """Tests for :py:class:`SbuildWorkflowData`."""

    def test_retry_delays(self) -> None:
        """A valid `retry_delays` field is accepted."""
        delays = ["30m", "1h", "2d", "1w"]
        data = SbuildWorkflowData(
            input=SbuildInput(source_artifact=42),
            target_distribution="debian:bookworm",
            architectures=["all"],
            retry_delays=delays,
        )
        self.assertEqual(
            data.model_dump(exclude_unset=True),
            {
                "input": {"source_artifact": 42},
                "target_distribution": "debian:bookworm",
                "architectures": ["all"],
                "retry_delays": delays,
            },
        )

    def test_retry_delays_bad_format(self) -> None:
        """Items in `retry_delays` must be integers followed by m/h/d/w."""
        with self.assertRaisesRegex(
            ValueError, r"String should match pattern.*input_value='2 weeks'"
        ):
            SbuildWorkflowData(
                input=SbuildInput(source_artifact=42),
                target_distribution="debian:bookworm",
                architectures=["all"],
                retry_delays=["2 weeks"],
            )


class RegressionTrackingWorkflowDataTests[WD: RegressionTrackingWorkflowData](
    TestCase, metaclass=ABCMeta
):
    """Common tests for models that handle regression tracking."""

    qa_suite_always_required = False

    @abstractmethod
    def make_data(self, **kwargs: Any) -> WD:
        """Make a data model of the appropriate type."""

    def test_update_qa_results(self) -> None:
        for value, expected in (
            (False, RegressionTrackingUpdateQAResults.NO),
            (True, RegressionTrackingUpdateQAResults.YES),
            ("no", RegressionTrackingUpdateQAResults.NO),
            ("yes", RegressionTrackingUpdateQAResults.YES),
            ("force", RegressionTrackingUpdateQAResults.FORCE),
        ):
            with self.subTest(value=value):
                self.assertEqual(
                    self.make_data(
                        qa_suite="sid@debian:suite",
                        reference_qa_results="sid@debian:qa-results",
                        update_qa_results=value,
                    ).update_qa_results,
                    expected,
                )

        with (
            self.subTest(value="nonsense"),
            self.assertRaisesRegex(
                ValueError,
                "'nonsense' is not a valid RegressionTrackingUpdateQAResults",
            ),
        ):
            self.make_data(update_qa_results="nonsense")

    def test_qa_results_consistency(self) -> None:
        """Test validation of regression-tracking options."""
        self.make_data()
        self.make_data(
            reference_prefix="reference-qa-result|",
            qa_suite="sid@debian:suite",
            reference_qa_results="sid@debian:qa-results",
            enable_regression_tracking=True,
        )
        self.make_data(
            qa_suite="sid@debian:suite",
            reference_qa_results="sid@debian:qa-results",
            update_qa_results=RegressionTrackingUpdateQAResults.YES,
        )
        self.make_data(
            qa_suite="sid@debian:suite",
            reference_qa_results="sid@debian:qa-results",
            update_qa_results=RegressionTrackingUpdateQAResults.FORCE,
        )

        kwargs: dict[str, Any]
        if not self.qa_suite_always_required:
            for kwargs in (
                {"enable_regression_tracking": True},
                {"update_qa_results": RegressionTrackingUpdateQAResults.YES},
                {"update_qa_results": RegressionTrackingUpdateQAResults.FORCE},
            ):
                with (
                    self.subTest(kwargs=str(kwargs)),
                    self.assertRaisesRegex(
                        ValueError,
                        '"qa_suite" is required if "enable_regression_tracking"'
                        ' or "update_qa_results" is set',
                    ),
                ):
                    self.make_data(**kwargs)
        for kwargs in (
            {
                "qa_suite": "sid@debian:suite",
                "enable_regression_tracking": True,
            },
            {
                "qa_suite": "sid@debian:suite",
                "update_qa_results": RegressionTrackingUpdateQAResults.YES,
            },
            {
                "qa_suite": "sid@debian:suite",
                "update_qa_results": RegressionTrackingUpdateQAResults.FORCE,
            },
        ):
            with (
                self.subTest(kwargs=str(kwargs)),
                self.assertRaisesRegex(
                    ValueError,
                    '"reference_qa_results" is required if'
                    ' "enable_regression_tracking" or "update_qa_results" is'
                    ' set',
                ),
            ):
                self.make_data(**kwargs)
        with self.assertRaisesRegex(
            ValueError,
            '"reference_prefix" is required if "enable_regression_tracking" '
            'is set',
        ):
            self.make_data(
                qa_suite="sid@debian:suite",
                reference_qa_results="sid@debian:qa-results",
                enable_regression_tracking=True,
            )


class PiupartsWorkflowDataTests(
    RegressionTrackingWorkflowDataTests[PiupartsWorkflowData]
):
    """Tests for :py:class:`PiupartsWorkflowData`."""

    @override
    def make_data(self, **kwargs: Any) -> PiupartsWorkflowData:
        """Make a `PiupartsWorkflowData`."""
        return PiupartsWorkflowData(
            source_artifact=1,
            binary_artifacts=LookupMultiple((2,)),
            vendor="debian",
            codename="sid",
            **kwargs,
        )


class AutopkgtestWorkflowDataTests(
    RegressionTrackingWorkflowDataTests[AutopkgtestWorkflowData]
):
    """Tests for :py:class:`AutopkgtestWorkflowData`."""

    @override
    def make_data(self, **kwargs: Any) -> AutopkgtestWorkflowData:
        """Make an `AutopkgtestWorkflowData`."""
        return AutopkgtestWorkflowData(
            source_artifact=1,
            binary_artifacts=LookupMultiple((2,)),
            vendor="debian",
            codename="sid",
            **kwargs,
        )


class ReverseDependenciesAutopkgtestWorkflowDataTests(
    RegressionTrackingWorkflowDataTests[
        ReverseDependenciesAutopkgtestWorkflowData
    ]
):
    """Tests for :py:class:`ReverseDependenciesAutopkgtestWorkflowData`."""

    qa_suite_always_required = True

    @override
    def make_data(
        self, **kwargs: Any
    ) -> ReverseDependenciesAutopkgtestWorkflowData:
        """Make a `ReverseDependenciesAutopkgtestWorkflowData`."""
        data: dict[str, Any] = {
            "source_artifact": 1,
            "binary_artifacts": LookupMultiple((2,)),
            "qa_suite": "sid@debian:suite",
            "vendor": "debian",
            "codename": "sid",
        }
        data.update(kwargs)
        return ReverseDependenciesAutopkgtestWorkflowData(**data)


class LintianWorkflowDataTests(
    RegressionTrackingWorkflowDataTests[LintianWorkflowData]
):
    """Tests for :py:class:`LintianWorkflowData`."""

    @override
    def make_data(self, **kwargs: Any) -> LintianWorkflowData:
        """Make a `LintianWorkflowData`."""
        return LintianWorkflowData(
            source_artifact=1,
            binary_artifacts=LookupMultiple((2,)),
            vendor="debian",
            codename="sid",
            **kwargs,
        )


class BlhcWorkflowDataTests(
    RegressionTrackingWorkflowDataTests[BlhcWorkflowData]
):
    """Tests for :py:class:`BlhcWorkflowData`."""

    @override
    def make_data(
        self,
        source_artifact: LookupSingle | None = 1,
        package_build_logs: LookupMultiple | None = LookupMultiple((2,)),
        **kwargs: Any,
    ) -> BlhcWorkflowData:
        """Make a `BlhcWorkflowData`."""
        return BlhcWorkflowData(
            source_artifact=source_artifact,
            package_build_logs=package_build_logs,
            vendor="debian",
            codename="sid",
            **kwargs,
        )

    def test_source_artifact(self) -> None:
        """Test validation of ``source_artifact``."""
        self.make_data(source_artifact=None)
        self.make_data(
            reference_prefix="reference-qa-result|",
            qa_suite="sid@debian:suite",
            reference_qa_results="sid@debian:qa-results",
            enable_regression_tracking=True,
        )

        with self.assertRaisesRegex(
            ValueError,
            '"source_artifact" is required if "enable_regression_tracking" is '
            'set',
        ):
            self.make_data(
                source_artifact=None,
                reference_prefix="reference-qa-result|",
                qa_suite="sid@debian:suite",
                reference_qa_results="sid@debian:qa-results",
                enable_regression_tracking=True,
            )

    def test_package_build_logs(self) -> None:
        self.make_data(
            binary_artifacts=LookupMultiple((1,)), package_build_logs=None
        )
        self.make_data(package_build_logs=LookupMultiple((2,)))

        error_msg = (
            "Exactly one of binary_artifacts or package_build_logs must be set"
        )
        with self.assertRaisesRegex(ValueError, error_msg):
            self.make_data(package_build_logs=None)
        with self.assertRaisesRegex(ValueError, error_msg):
            self.make_data(
                binary_artifacts=LookupMultiple((1,)),
                package_build_logs=LookupMultiple((2,)),
            )


class QAWorkflowDataTests(RegressionTrackingWorkflowDataTests[QAWorkflowData]):
    """Tests for :py:class:`QAWorkflowData`."""

    @override
    def make_data(self, **kwargs: Any) -> QAWorkflowData:
        """Make a `QAWorkflowData`."""
        return QAWorkflowData.model_validate(
            {
                "source_artifact": 1,
                "binary_artifacts": LookupMultiple((2,)),
                "vendor": "debian",
                "codename": "sid",
                **kwargs,
            },
        )

    def test_reverse_dependencies_autopkgtest_consistency(self) -> None:
        """Test reverse-dependencies-autopkgtest validation."""
        self.make_data()
        self.make_data(
            qa_suite="sid@debian:suite",
            enable_reverse_dependencies_autopkgtest=True,
        )

        with self.assertRaisesRegex(
            ValueError,
            '"qa_suite" is required if '
            '"enable_reverse_dependencies_autopkgtest" is set',
        ):
            self.make_data(enable_reverse_dependencies_autopkgtest=True)

    def test_enable_debdiff_consistency(self) -> None:
        """Test enable_debdiff_consistency validation."""
        self.make_data()
        self.make_data(qa_suite="sid@debian:suite", enable_debdiff=True)

        self.make_data()

        with self.assertRaisesRegex(
            ValueError,
            '"qa_suite" is required if "enable_debdiff" is set',
        ):
            self.make_data(enable_debdiff=True)

    def test_default_fail_on(self) -> None:
        """Test default for ``fail_on``."""
        self.assertEqual(self.make_data().fail_on, QAFailOn.FAILURE)
        self.assertEqual(
            self.make_data(
                reference_prefix="reference-qa-result|",
                qa_suite="sid@debian:suite",
                reference_qa_results="sid@debian:qa-results",
                enable_regression_tracking=True,
            ).fail_on,
            QAFailOn.REGRESSION,
        )


class DebianPipelineWorkflowDataTests(TestCase):
    """Tests for :py:class:`DebianPipelineWorkflowData`."""

    def test_reverse_dependencies_autopkgtest_consistency(self) -> None:
        """Test reverse-dependencies-autopkgtest validation."""
        DebianPipelineWorkflowData(
            source_artifact=1, vendor="debian", codename="sid"
        )
        DebianPipelineWorkflowData(
            source_artifact=1,
            vendor="debian",
            codename="sid",
            qa_suite="sid@debian:suite",
            enable_reverse_dependencies_autopkgtest=True,
        )

        with self.assertRaisesRegex(
            ValueError,
            '"qa_suite" is required if '
            '"enable_reverse_dependencies_autopkgtest" is set',
        ):
            DebianPipelineWorkflowData(
                source_artifact=1,
                vendor="debian",
                codename="sid",
                enable_reverse_dependencies_autopkgtest=True,
            )

    def test_enable_debdiff_consistency(self) -> None:
        """Test enable_debdiff validator."""
        DebianPipelineWorkflowData(
            source_artifact=1, vendor="debian", codename="sid"
        )

        with self.assertRaisesRegex(
            ValueError,
            '"qa_suite" is required if "enable_debdiff" is set',
        ):
            DebianPipelineWorkflowData(
                source_artifact=1,
                enable_debdiff=True,
                vendor="debian",
                codename="sid",
            )

    def test_regression_tracking_consistency(self) -> None:
        """Test regression-tracking validation."""
        DebianPipelineWorkflowData(
            source_artifact=1, vendor="debian", codename="sid"
        )
        DebianPipelineWorkflowData(
            source_artifact=1,
            vendor="debian",
            codename="sid",
            qa_suite="sid@debian:suite",
            regression_tracking_qa_results="sid@debian:qa-results",
            enable_regression_tracking=True,
        )

        with self.assertRaisesRegex(
            ValueError,
            '"qa_suite" is required if "enable_regression_tracking" is set',
        ):
            DebianPipelineWorkflowData(
                source_artifact=1,
                vendor="debian",
                codename="sid",
                enable_regression_tracking=True,
            )
        with self.assertRaisesRegex(
            ValueError,
            '"regression_tracking_qa_results" is required if '
            '"enable_regression_tracking" is set',
        ):
            DebianPipelineWorkflowData(
                source_artifact=1,
                vendor="debian",
                codename="sid",
                qa_suite="sid@debian:suite",
                enable_regression_tracking=True,
            )


class PackagePublishWorkflowDataTests(TestCase):
    """Tests for :py:class:`PackagePublishWorkflowData`."""

    def test_input_validation(self) -> None:
        """Test LintianInput validation."""
        PackagePublishWorkflowData(
            source_artifact=1, target_suite="bookworm@debian:suite"
        )
        PackagePublishWorkflowData(
            binary_artifacts=LookupMultiple((1,)),
            target_suite="bookworm@debian:suite",
        )
        PackagePublishWorkflowData(
            binary_artifacts=LookupMultiple((1, 2)),
            target_suite="bookworm@debian:suite",
        )

        error_msg = '"source_artifact" or "binary_artifacts" must be set'
        with self.assertRaisesRegex(ValueError, error_msg):
            PackagePublishWorkflowData(target_suite="bookworm@debian:suite")
        with self.assertRaisesRegex(ValueError, error_msg):
            PackagePublishWorkflowData(
                binary_artifacts=LookupMultiple(()),
                target_suite="bookworm@debian:suite",
            )


class ChildWorkspaceDataTests(TestCase):
    """Tests for :py:class:`ChildWorkspaceData`."""

    def test_validate_prefix(self) -> None:
        """Test prefix validation."""
        for name in [None, "test-test"]:
            with self.subTest(name=name):
                ChildWorkspaceData(prefix=name, suffix="suffix")

        for name in ["0test", "test@test", ".test", "+test"]:
            with (
                self.subTest(name=name),
                self.assertRaisesRegex(
                    ValueError, r"\bprefix\n.*String should match pattern"
                ),
            ):
                ChildWorkspaceData(prefix=name, suffix="suffix")

    def test_validate_suffix(self) -> None:
        """Test suffix validation."""
        for name in ["test-test"]:
            with self.subTest(name=name):
                ChildWorkspaceData(suffix=name)

        for name in ["0test", "test@test", ".test", "+test"]:
            with (
                self.subTest(name=name),
                self.assertRaisesRegex(
                    ValueError, r"\bsuffix\n.*String should match pattern"
                ),
            ):
                ChildWorkspaceData(suffix=name)


# Avoid running tests from common base class.
del RegressionTrackingWorkflowDataTests



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Custom routers for the server API."""

from typing import override

from rest_framework.routers import DefaultRouter
from rest_framework.viewsets import ViewSetMixin


class APIRouter(DefaultRouter):
    """A router that can handle looking up objects by multiple fields."""

    @override
    def get_lookup_regex(
        self, viewset: type[ViewSetMixin], lookup_prefix: str = ""
    ) -> str:
        if hasattr(viewset, "get_lookup_regex"):
            assert getattr(self, "_use_regex")
            lookup_regex = viewset.get_lookup_regex(lookup_prefix=lookup_prefix)
            assert isinstance(lookup_regex, str)
            return lookup_regex
        else:
            return super().get_lookup_regex(
                viewset, lookup_prefix=lookup_prefix
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the debusine Cli task configuration commands."""

import os
from functools import partial
from pathlib import Path
from typing import Any, ClassVar, override
from unittest import mock
from unittest.mock import Mock

import yaml

from debusine.artifacts.models import DebusineTaskConfiguration, TaskTypes
from debusine.client.commands.task_config import Pull
from debusine.client.commands.tests.base import BaseCliTests
from debusine.client.models import (
    TaskConfigurationCollection,
    TaskConfigurationCollectionUpdateResults,
)
from debusine.client.task_configuration import RemoteTaskConfigurationRepository


class CliTaskConfigPullPushTests(BaseCliTests):
    """Tests for Cli task-config-pull and -push."""

    config1: ClassVar[DebusineTaskConfiguration]
    config2: ClassVar[DebusineTaskConfiguration]
    template: ClassVar[DebusineTaskConfiguration]

    @override
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.config1 = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER, task_name="noop"
        )
        cls.config2 = DebusineTaskConfiguration(
            task_type=TaskTypes.WORKER,
            task_name="noop",
            context="context",
            subject="subject",
        )
        cls.template = DebusineTaskConfiguration(
            template="template", comment="this is a template"
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        # Working directory for the local collection
        self.workdir = self.create_temporary_directory()
        # Mocked debusine object
        build_debusine = self.patch_build_debusine_object()
        self.debusine = build_debusine.return_value

    def manifest(
        self,
        workspace: str = "System",
        pk: int = 42,
        name: str = "default",
        data: dict[str, Any] | None = None,
    ) -> RemoteTaskConfigurationRepository.Manifest:
        """Create a manifest."""
        return RemoteTaskConfigurationRepository.Manifest(
            workspace=workspace,
            collection=TaskConfigurationCollection(
                id=pk, name=name, data=data or {}
            ),
        )

    def make_remote_repo(
        self, workspace: str = "System", pk: int = 42, name: str = "default"
    ) -> RemoteTaskConfigurationRepository:
        """Create a TaskConfigurationRepository with manifest."""
        return RemoteTaskConfigurationRepository(
            self.manifest(workspace=workspace, pk=pk, name=name)
        )

    def mock_fetch(self, repo: RemoteTaskConfigurationRepository) -> Mock:
        """Mock push_task_configuration_collection with a plausible return."""
        fetch = Mock(return_value=repo)
        self.debusine.fetch_task_configuration_collection = fetch
        return fetch

    def mock_push(
        self,
        added: int = 1,
        updated: int = 2,
        removed: int = 3,
        unchanged: int = 4,
    ) -> Mock:
        """Mock push_task_configuration_collection with a plausible return."""
        results = TaskConfigurationCollectionUpdateResults(
            added=added, updated=updated, removed=removed, unchanged=unchanged
        )
        push = Mock(return_value=results)
        self.debusine.push_task_configuration_collection = push
        return push

    def write_file(self, path: Path, data: dict[str, Any]) -> None:
        """Write out a yaml file in workdir."""
        with (self.workdir / path).open("wt") as fd:
            yaml.safe_dump(data, fd)

    def workdir_files(self) -> list[Path]:
        """List files present in workdir."""
        res: list[Path] = []
        for cur_root, _dirs, files in self.workdir.walk():
            res += [
                (cur_root / name).relative_to(self.workdir) for name in files
            ]
        return res

    def assertFileContents(self, path: Path, expected: dict[str, Any]) -> None:
        """Check that path points to a YAML file with the given contents."""
        with (self.workdir / path).open() as fd:
            actual = yaml.safe_load(fd)
        self.assertEqual(actual, expected)

    def test_pull_default_to_current_dir(self) -> None:
        orig_cwd = os.getcwd()
        try:
            os.chdir(self.workdir)
            command = self.create_command(["task-config", "pull"])
            assert isinstance(command, Pull)
            self.assertEqual(command.args.workdir, Path.cwd())
            self.assertEqual(command.workspace, "developers")
            self.assertEqual(command.collection, "default")
        finally:
            os.chdir(orig_cwd)

    def test_pull_workdir(self) -> None:
        command = self.create_command(
            ["task-config", "pull", "--workdir", self.workdir.as_posix()]
        )
        assert isinstance(command, Pull)
        self.assertEqual(command.args.workdir, self.workdir)
        self.assertEqual(command.workspace, "developers")
        self.assertEqual(command.collection, "default")

    def test_pull_workspace(self) -> None:
        command = self.create_command(
            [
                "task-config",
                "pull",
                "--workdir",
                self.workdir.as_posix(),
                "--workspace=workspace",
            ]
        )
        assert isinstance(command, Pull)
        self.assertEqual(command.args.workdir, self.workdir)
        self.assertEqual(command.workspace, "workspace")
        self.assertEqual(command.collection, "default")

    def test_pull_collection_name(self) -> None:
        command = self.create_command(
            [
                "task-config",
                "pull",
                "--workdir",
                self.workdir.as_posix(),
                "--workspace=workspace",
                "collection",
            ]
        )
        assert isinstance(command, Pull)
        self.assertEqual(command.args.workdir, self.workdir)
        self.assertEqual(command.workspace, "workspace")
        self.assertEqual(command.collection, "collection")

    def test_pull_workspace_must_match_checkout(self) -> None:
        self.write_file(
            Path("MANIFEST"),
            self.manifest(workspace="workspace", name="name").model_dump(),
        )
        stderr, stdout = self.capture_output(
            partial(
                self.create_command,
                [
                    "task-config",
                    "pull",
                    "--workdir",
                    self.workdir.as_posix(),
                    "--workspace=other_workspace",
                    "name",
                ],
            ),
            assert_system_exit_code=3,
        )
        self.assertEqual(
            stderr,
            "--workspace=other_workspace is provided,"
            " but the repository checked out refers to"
            " workspace 'workspace'\n",
        )
        self.assertEqual(stdout, "")

    def test_pull_collection_must_match_checkout_ok(self) -> None:
        self.write_file(
            Path("MANIFEST"),
            self.manifest(workspace="workspace", name="name").model_dump(),
        )
        stderr, stdout = self.capture_output(
            partial(
                self.create_command,
                [
                    "task-config",
                    "pull",
                    "--workdir",
                    self.workdir.as_posix(),
                    "--workspace=workspace",
                    "name",
                ],
            ),
        )
        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")

    def test_pull_collection_must_match_checkout_fail(self) -> None:
        self.write_file(
            Path("MANIFEST"),
            self.manifest(workspace="workspace", name="name").model_dump(),
        )
        stderr, stdout = self.capture_output(
            partial(
                self.create_command,
                [
                    "task-config",
                    "pull",
                    "--workdir",
                    self.workdir.as_posix(),
                    "--workspace=workspace",
                    "other_name",
                ],
            ),
            assert_system_exit_code=3,
        )
        self.assertEqual(
            stderr,
            "--collection=other_name is provided,"
            " but the repository checked out refers to collection 'name'\n",
        )
        self.assertEqual(stdout, "")

    def test_pull_new_repo(self) -> None:
        server_repo = self.make_remote_repo()
        server_repo.entries += [self.config1]
        self.mock_fetch(repo=server_repo)
        cli = self.create_cli(
            [
                "task-config",
                "pull",
                "--workspace=System",
                "--workdir",
                self.workdir.as_posix(),
            ]
        )
        with self.assertLogs("debusine.client.tests") as log:
            stderr, stdout = self.capture_output(cli.execute)

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        self.assertCountEqual(
            [x.removeprefix("INFO:debusine.client.tests:") for x in log.output],
            [
                "new.yaml: new file with 1 entries",
                "1 added, 0 updated, 0 deleted, 0 unchanged",
            ],
        )
        self.assertCountEqual(
            self.workdir_files(),
            [path_manifest := Path("MANIFEST"), path_new := Path("new.yaml")],
        )

        self.assertFileContents(
            path_manifest, server_repo.manifest.model_dump()
        )
        self.assertFileContents(path_new, self.config1.model_dump())

    def test_pull_existing_repo(self) -> None:
        server_repo = self.make_remote_repo()
        server_repo.entries += [self.config1, self.template]
        self.mock_fetch(repo=server_repo)
        self.write_file(
            path_manifest := Path("MANIFEST"), server_repo.manifest.model_dump()
        )
        self.write_file(
            path_config1 := Path("config.yaml"), self.config1.model_dump()
        )
        self.write_file(
            path_template := Path("template.yaml"), self.template.model_dump()
        )
        cli = self.create_cli(
            ["task-config", "pull", "--workdir", self.workdir.as_posix()]
        )
        with self.assertLogs("debusine.client.tests") as log:
            stderr, stdout = self.capture_output(cli.execute)

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        self.assertCountEqual(
            [x.removeprefix("INFO:debusine.client.tests:") for x in log.output],
            [
                "0 added, 0 updated, 0 deleted, 2 unchanged",
            ],
        )
        self.assertCountEqual(
            self.workdir_files(),
            [
                path_manifest,
                path_template,
                path_config1,
            ],
        )

        self.assertFileContents(
            path_manifest, server_repo.manifest.model_dump()
        )
        self.assertFileContents(path_template, self.template.model_dump())
        self.assertFileContents(path_config1, self.config1.model_dump())

    def test_pull_repo_manifest_mismatch(self) -> None:
        server_repo = self.make_remote_repo()
        server_repo.entries += [self.config2]
        self.mock_fetch(repo=server_repo)
        local_manifest = server_repo.manifest.model_dump()
        local_manifest["workspace"] = "other"
        self.write_file(path_manifest := Path("MANIFEST"), local_manifest)
        self.write_file(
            path_config1 := Path("config.yaml"), self.config1.model_dump()
        )
        cli = self.create_cli(
            [
                "task-config",
                "pull",
                "--workdir",
                self.workdir.as_posix(),
            ]
        )
        stderr, stdout = self.capture_output(
            cli.execute, assert_system_exit_code=3
        )

        self.assertEqual(
            stderr,
            "repo to pull refers to collection System/default (42)"
            " while the checkout has collection other/default (42)\n",
        )
        self.assertEqual(stdout, "")
        self.assertCountEqual(
            self.workdir_files(),
            [
                path_manifest,
                path_config1,
            ],
        )

    def test_push_default_to_current_dir(self) -> None:
        command = self.create_command(["task-config", "push"])
        self.assertEqual(command.args.workdir, Path.cwd())
        self.assertFalse(command.args.dry_run)
        self.assertFalse(command.args.force)

    def test_push_workdir(self) -> None:
        command = self.create_command(
            ["task-config", "push", "--workdir", self.workdir.as_posix()]
        )
        self.assertEqual(command.args.workdir, self.workdir)
        self.assertFalse(command.args.dry_run)
        self.assertFalse(command.args.force)

    def test_push_dry_run(self) -> None:
        command = self.create_command(
            [
                "task-config",
                "push",
                "--workdir",
                self.workdir.as_posix(),
                "--dry-run",
            ]
        )
        self.assertEqual(command.args.workdir, self.workdir)
        self.assertTrue(command.args.dry_run)
        self.assertFalse(command.args.force)

    def test_push_force(self) -> None:
        command = self.create_command(
            [
                "task-config",
                "push",
                "--workdir",
                self.workdir.as_posix(),
                "--force",
            ]
        )
        self.assertEqual(command.args.workdir, self.workdir)
        self.assertFalse(command.args.dry_run)
        self.assertTrue(command.args.force)

    def test_push_no_workdir(self) -> None:
        path = self.workdir / "does-not-exist"
        cli = self.create_cli(
            ["task-config", "push", "--workdir", path.as_posix()]
        )
        stderr, stdout = self.capture_output(
            cli.execute, assert_system_exit_code=3
        )
        self.assertEqual(stderr, f"{path} is not a repository\n")
        self.assertEqual(stdout, "")

    def test_push_no_manifest(self) -> None:
        cli = self.create_cli(
            ["task-config", "push", "--workdir", self.workdir.as_posix()]
        )
        stderr, stdout = self.capture_output(
            cli.execute, assert_system_exit_code=3
        )
        self.assertEqual(stderr, f"{self.workdir} is not a repository\n")
        self.assertEqual(stdout, "")

    def test_push_repo(self) -> None:
        server_repo = self.make_remote_repo()
        self.mock_fetch(repo=server_repo)
        push_method = self.mock_push()
        self.write_file(Path("MANIFEST"), self.manifest().model_dump())
        self.write_file(Path("config.yaml"), self.config1.model_dump())
        cli = self.create_cli(
            ["task-config", "push", "--workdir", self.workdir.as_posix()]
        )
        with self.assertLogs("debusine.client.tests") as log:
            stderr, stdout = self.capture_output(cli.execute)

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        self.assertCountEqual(
            [x.removeprefix("INFO:debusine.client.tests:") for x in log.output],
            [
                "Pushing data to server...",
                "1 added, 2 updated, 3 removed, 4 unchanged",
            ],
        )

        repo = push_method.call_args.kwargs["repo"]
        self.assertEqual(repo.manifest, self.manifest())
        self.assertEqual(
            list(repo.iter_entries()),
            [
                self.config1.model_copy(
                    update={"path": "config.yaml", "position": 0}
                )
            ],
        )
        self.assertFalse(push_method.call_args.kwargs["dry_run"])

    def test_push_repo_dry_run(self) -> None:
        server_repo = self.make_remote_repo()
        self.mock_fetch(repo=server_repo)
        push_method = self.mock_push()
        self.write_file(Path("MANIFEST"), self.manifest().model_dump())
        self.write_file(Path("config.yaml"), self.config1.model_dump())
        cli = self.create_cli(
            [
                "task-config",
                "push",
                "--workdir",
                self.workdir.as_posix(),
                "--dry-run",
            ]
        )
        with self.assertLogs("debusine.client.tests") as log:
            stderr, stdout = self.capture_output(cli.execute)

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        self.assertCountEqual(
            [x.removeprefix("INFO:debusine.client.tests:") for x in log.output],
            [
                "Pushing data to server (dry run)...",
                "1 added, 2 updated, 3 removed, 4 unchanged",
            ],
        )

        repo = push_method.call_args.kwargs["repo"]
        self.assertEqual(repo.manifest, self.manifest())
        self.assertEqual(
            list(repo.iter_entries()),
            [
                self.config1.model_copy(
                    update={"path": "config.yaml", "position": 0}
                )
            ],
        )
        self.assertTrue(push_method.call_args.kwargs["dry_run"])

    def test_push_repo_dirty(self) -> None:
        self.write_file(Path("MANIFEST"), self.manifest().model_dump())
        cli = self.create_cli(
            ["task-config", "push", "--workdir", self.workdir.as_posix()]
        )
        with mock.patch(
            "debusine.client.task_configuration"
            ".LocalTaskConfigurationRepository.is_dirty",
            return_value=True,
        ):
            stderr, stdout = self.capture_output(
                cli.execute, assert_system_exit_code=3
            )
        self.assertEqual(
            stderr,
            f"{self.workdir} has uncommitted changes:"
            " please commit them before pushing\n",
        )
        self.assertEqual(stdout, "")

    def test_push_repo_dirty_force(self) -> None:
        self.write_file(Path("MANIFEST"), self.manifest().model_dump())
        fetch_method = self.mock_fetch(repo=self.make_remote_repo())
        push_method = self.mock_push()
        cli = self.create_cli(
            [
                "task-config",
                "push",
                "--workdir",
                self.workdir.as_posix(),
                "--force",
            ]
        )
        with (
            mock.patch(
                "debusine.client.task_configuration"
                ".LocalTaskConfigurationRepository.is_dirty",
                return_value=True,
            ),
            self.assertLogs("debusine.client.tests") as log,
        ):
            stderr, stdout = self.capture_output(cli.execute)
        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        self.assertCountEqual(
            [
                x.removeprefix("WARNING:debusine.client.tests:").removeprefix(
                    "INFO:debusine.client.tests:"
                )
                for x in log.output
            ],
            [
                f"{self.workdir} has uncommitted changes:"
                " please commit them before pushing",
                "Pushing data to server...",
                "1 added, 2 updated, 3 removed, 4 unchanged",
            ],
        )
        fetch_method.assert_called()
        push_method.assert_called()

    def test_push_git_no_commit_on_server(self) -> None:
        local_commit = "0" * 40
        server_repo = self.make_remote_repo()
        self.mock_fetch(repo=server_repo)
        push_method = self.mock_push()
        self.write_file(Path("MANIFEST"), self.manifest().model_dump())
        with mock.patch(
            "debusine.client.task_configuration"
            ".LocalTaskConfigurationRepository.git_commit",
            return_value=local_commit,
        ):
            cli = self.create_cli(
                [
                    "task-config",
                    "push",
                    "--workdir",
                    self.workdir.as_posix(),
                ]
            )
            stderr, stdout = self.capture_output(cli.execute)

        self.assertEqual(
            stderr,
            (
                "Pushing data to server...\n"
                "1 added, 2 updated, 3 removed, 4 unchanged\n"
            ),
        )
        self.assertEqual(stdout, "")

        repo = push_method.call_args.kwargs["repo"]
        self.assertEqual(
            repo.manifest, self.manifest(data={"git_commit": local_commit})
        )

    def test_push_git_does_not_have_previous_commit(self) -> None:
        local_commit = "1" * 40
        server_commit = "0" * 40
        server_repo = self.make_remote_repo()
        server_repo.manifest.collection.data["git_commit"] = server_commit
        fetch_method = self.mock_fetch(repo=server_repo)
        push_method = self.mock_push()

        self.write_file(Path("MANIFEST"), self.manifest().model_dump())
        with (
            mock.patch(
                "debusine.client.task_configuration"
                ".LocalTaskConfigurationRepository.git_commit",
                return_value=local_commit,
            ),
            mock.patch(
                "debusine.client.task_configuration"
                ".LocalTaskConfigurationRepository.has_commit",
                return_value=False,
            ),
        ):
            cli = self.create_cli(
                [
                    "task-config",
                    "push",
                    "--workdir",
                    self.workdir.as_posix(),
                ]
            )
            stderr, stdout = self.capture_output(
                cli.execute, assert_system_exit_code=3
            )

        self.assertEqual(
            stderr,
            "server collection was pushed from commit"
            f" {server_commit} which is not known to {self.workdir}\n",
        )
        self.assertEqual(stdout, "")

        fetch_method.assert_called()
        push_method.assert_not_called()

    def test_push_git_does_not_have_previous_commit_force(self) -> None:
        local_commit = "1" * 40
        server_commit = "0" * 40
        server_repo = self.make_remote_repo()
        server_repo.manifest.collection.data["git_commit"] = server_commit
        fetch_method = self.mock_fetch(repo=server_repo)
        push_method = self.mock_push()

        self.write_file(Path("MANIFEST"), self.manifest().model_dump())
        with (
            mock.patch(
                "debusine.client.task_configuration"
                ".LocalTaskConfigurationRepository.git_commit",
                return_value=local_commit,
            ),
            mock.patch(
                "debusine.client.task_configuration"
                ".LocalTaskConfigurationRepository.has_commit",
                return_value=False,
            ),
            self.assertLogs("debusine.client.tests") as log,
        ):
            cli = self.create_cli(
                [
                    "task-config",
                    "push",
                    "--workdir",
                    self.workdir.as_posix(),
                    "--force",
                ]
            )
            stderr, stdout = self.capture_output(cli.execute)

        self.assertCountEqual(
            [
                x.removeprefix("WARNING:debusine.client.tests:").removeprefix(
                    "INFO:debusine.client.tests:"
                )
                for x in log.output
            ],
            [
                "server collection was pushed from commit"
                f" {server_commit} which is not known to {self.workdir}",
                "Pushing data to server...",
                "1 added, 2 updated, 3 removed, 4 unchanged",
            ],
        )

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")

        fetch_method.assert_called()
        push_method.assert_called()

    def test_push_git_has_previous_commit(self) -> None:
        local_commit = "0" * 40
        server_commit = "1" * 40
        server_repo = self.make_remote_repo()
        server_repo.manifest.collection.data["git_commit"] = server_commit
        self.mock_fetch(repo=server_repo)
        push_method = self.mock_push()

        self.write_file(Path("MANIFEST"), self.manifest().model_dump())
        with (
            mock.patch(
                "debusine.client.task_configuration"
                ".LocalTaskConfigurationRepository.git_commit",
                return_value=local_commit,
            ),
            mock.patch(
                "debusine.client.task_configuration"
                ".LocalTaskConfigurationRepository.has_commit",
                return_value=True,
            ),
        ):
            cli = self.create_cli(
                [
                    "task-config",
                    "push",
                    "--workdir",
                    self.workdir.as_posix(),
                ]
            )
            stderr, stdout = self.capture_output(cli.execute)

        self.assertEqual(
            stderr,
            (
                "Pushing data to server...\n"
                "1 added, 2 updated, 3 removed, 4 unchanged\n"
            ),
        )
        self.assertEqual(stdout, "")

        repo = push_method.call_args.kwargs["repo"]
        self.assertEqual(
            repo.manifest, self.manifest(data={"git_commit": local_commit})
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Transparently add scopes to URLs."""

import contextlib
import re
import sys
import types
from collections.abc import Generator

from django.conf import settings
from django.urls import get_urlconf, set_urlconf

from debusine.project.urls import make_urlpatterns

scope_prefix_re = re.compile(r'^/([^/]+)(/|$)')


def get_scope_urlconf(scope_name: str) -> str:
    """Return a per-scope urlconf."""
    if scope_name == settings.DEBUSINE_DEFAULT_SCOPE:
        return "debusine.project.urls"

    module_name = f"debusine.server._urlconfs.{scope_name}"
    if module_name not in sys.modules:
        mod = types.ModuleType(scope_name, f"URLConf for scope {scope_name!r}")
        setattr(mod, "urlpatterns", make_urlpatterns(scope_name))
        sys.modules[module_name] = mod

    return module_name


@contextlib.contextmanager
def urlconf_scope(scope_name: str) -> Generator[None]:
    """Temporarily set the scope for reverse URL resolution."""
    orig = get_urlconf()
    try:
        # Note that this does the right thing also with threads and channels,
        # since Django's set_urlconf is backed by asgiref.local.Local.
        set_urlconf(get_scope_urlconf(scope_name))
        yield
    finally:
        set_urlconf(orig)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to manage users."""

from typing import Any, NoReturn, override

from django.contrib.auth import get_user_model
from django.core.management import CommandError, CommandParser

from debusine.django.management.debusine_base_command import DebusineBaseCommand


class Command(DebusineBaseCommand):
    """Command to manage users."""

    help = "Manage users"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        subparsers = parser.add_subparsers(dest="action", required=True)

        # change email
        change_email = subparsers.add_parser("change-email")
        change_email.add_argument("username", help="Username to modify")
        change_email.add_argument("email", help="New email to set")

    @override
    def handle(self, *args: Any, **options: Any) -> None:
        """Manage the user."""
        if options["action"] == "change-email":
            self._change_email(options["username"], options["email"])
        else:  # pragma: no cover
            pass  # pragma: no cover

    def _change_email(self, username: str, email: str) -> NoReturn:
        try:
            user = get_user_model().objects.get(
                username=username, is_system=False
            )
        except get_user_model().DoesNotExist:
            raise CommandError(f'Username "{username}" not found', returncode=3)

        user.email = email
        user.save()

        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the task test helpers."""

from debusine.artifacts.models import WorkRequestResults
from debusine.tasks import Noop
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase


class HelpersTests(ExternalTaskHelperMixin[Noop], TestCase):
    """Test the external task test helpers."""

    def test_set_current_does_not_manage_twice(self) -> None:
        # Task is managed on creation
        task1 = self.task
        assert task1 in self.tasks
        self.task_reset_current()

        # A new task is also managed on creation
        self.task_data["result"] = WorkRequestResults.FAILURE
        task2 = self.task
        assert task2 in self.tasks

        assert task1 != task2

        with self.spy(
            "debusine.tasks.tests.helpers.TaskTestCase.task_manage"
        ) as task_manage:
            self.task_set_current(task1)

        self.assertIs(self.task, task1)
        task_manage.assert_not_called()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine workflow templates views."""

from typing import cast, override

from django.db.models import QuerySet
from django.shortcuts import get_object_or_404

from debusine.db.models.work_requests import (
    WorkflowTemplate,
    WorkflowTemplateQuerySet,
)
from debusine.web.views.base import DetailViewBase, WorkspaceView
from debusine.web.views.places import Place


class WorkflowTemplateDetailView(
    WorkspaceView, DetailViewBase[WorkflowTemplate]
):
    """Display details about the workflow template."""

    model = WorkflowTemplate
    template_name = "web/workflow-template-detail.html"
    context_object_name = "workflow_template"

    @override
    def get_queryset(self) -> QuerySet[WorkflowTemplate]:
        """Filter workflow templates to the current workspace."""
        return cast(
            WorkflowTemplateQuerySet[WorkflowTemplate], super().get_queryset()
        ).in_current_workspace()

    @override
    def get_object(
        self,
        queryset: QuerySet[WorkflowTemplate, WorkflowTemplate] | None = None,
    ) -> WorkflowTemplate:
        """Return the workflow template object to show."""
        assert queryset is None
        queryset = self.get_queryset()

        workflow_template = get_object_or_404(
            queryset, name=self.kwargs["name"]
        )
        self.enforce(workflow_template.can_display)
        return workflow_template

    @override
    def get_place(self) -> Place:
        return self.object.ui(self.request).place



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the database aware task input value models."""

from typing import ClassVar, override

from debusine.artifacts.models import CollectionCategory
from debusine.client.models import LookupResultType
from debusine.db.models import Artifact, Collection, CollectionItem
from debusine.db.models.lookups import (
    LookupResultArtifact,
    LookupResultBare,
    LookupResultCollection,
)
from debusine.db.tasks.models import (
    DBArtifactInfo,
    DBBaseTaskInputArtifact,
    DBInputCollectionSingle,
    DBInputLookupResult,
)
from debusine.tasks.models import LookupSingle
from debusine.test.django import TestCase


class DBBaseTaskInputArtifactTests(TestCase):
    """Tests for :py:class:`DBBaseTaskInputArtifact`."""

    artifact: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.artifact = cls.playground.create_minimal_source_package_artifact()

    def test_from_artifact(self) -> None:
        value = DBBaseTaskInputArtifact.from_artifact(self.artifact)
        self.assertEqual(value.artifact_id, self.artifact.id)
        self.assertEqual(value.category, self.artifact.category)
        self.assertEqual(value.data, self.artifact.create_data())
        self.assertIsNotNone(value._artifact)
        with self.assertNumQueries(0):
            self.assertEqual(value.artifact, self.artifact)

    def test_cache_miss(self) -> None:
        value = DBBaseTaskInputArtifact.from_artifact(self.artifact)
        value._artifact = None
        with self.assertNumQueries(1):
            self.assertEqual(value.artifact, self.artifact)


class DBArtifactInfoTests(TestCase):
    """Tests for :py:class:`DBArtifactInfo`."""

    artifact: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.artifact = cls.playground.create_minimal_source_package_artifact()

    def test_from_artifact(self) -> None:
        value = DBArtifactInfo.from_artifact(
            self.artifact, label="label", lookup="lookup"
        )
        self.assertEqual(value.label, "label")
        self.assertEqual(value.lookup, "lookup")
        self.assertEqual(value.artifact_id, self.artifact.id)
        self.assertEqual(value.category, self.artifact.category)
        self.assertEqual(value.data, self.artifact.create_data())
        self.assertIsNotNone(value._artifact)
        with self.assertNumQueries(0):
            self.assertEqual(value.artifact, self.artifact)

    def test_cache_miss(self) -> None:
        value = DBArtifactInfo.from_artifact(
            self.artifact, label="label", lookup="lookup"
        )
        value._artifact = None
        with self.assertNumQueries(1):
            self.assertEqual(value.artifact, self.artifact)


class DBInputCollectionSingleTests(TestCase):
    """Tests for :py:class:`DBInputCollectionSingle`."""

    collection: ClassVar[Collection]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.collection = cls.playground.create_collection(
            "test", CollectionCategory.TEST
        )

    def test_from_collection(self) -> None:
        value = DBInputCollectionSingle.from_collection(
            self.collection, label="label", lookup="lookup"
        )
        self.assertEqual(value.label, "label")
        self.assertEqual(value.lookup, "lookup")
        self.assertEqual(value.collection_id, self.collection.id)
        self.assertEqual(value.scope_name, self.collection.workspace.scope.name)
        self.assertEqual(value.workspace_name, self.collection.workspace.name)
        self.assertEqual(value.category, self.collection.category)
        self.assertEqual(value.name, self.collection.name)
        self.assertEqual(value.data, self.collection.data)
        self.assertIsNotNone(value._collection)
        with self.assertNumQueries(0):
            self.assertEqual(value.collection, self.collection)

    def test_cache_miss(self) -> None:
        value = DBInputCollectionSingle.from_collection(
            self.collection, label="label", lookup="lookup"
        )
        value._collection = None
        with self.assertNumQueries(1):
            self.assertEqual(value.collection, self.collection)


class DBInputLookupResultTests(TestCase):
    """Tests for :py:class:`DBInputLookupResult`."""

    collection: ClassVar[Collection]
    collection_lookup: ClassVar[LookupSingle]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.collection = cls.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        cls.collection_lookup = f"{cls.collection.pk}@collections"

    def make_lookup_result_bare(self) -> LookupResultBare:
        collection_item = self.playground.create_bare_data_item(
            parent_collection=self.collection, name="test"
        )
        return LookupResultBare(
            lookup="lookup",
            parent_collection_lookup=self.collection_lookup,
            collection_item=collection_item,
        )

    def make_lookup_result_artifact(self) -> LookupResultArtifact:
        artifact, _ = self.playground.create_artifact()
        collection_item = CollectionItem.objects.create_from_artifact(
            artifact,
            parent_collection=self.collection,
            name="test",
            data={},
            created_by_user=self.playground.get_default_user(),
        )
        return LookupResultArtifact(
            lookup="lookup",
            parent_collection_lookup=self.collection_lookup,
            collection_item=collection_item,
            artifact=artifact,
        )

    def make_lookup_result_collection(self) -> LookupResultCollection:
        sub_collection = self.playground.create_collection(
            name="sub", category=CollectionCategory.TEST
        )
        collection_item = CollectionItem.objects.create_from_collection(
            sub_collection,
            parent_collection=self.collection,
            name="test",
            data={},
            created_by_user=self.playground.get_default_user(),
        )
        return LookupResultCollection(
            lookup="lookup",
            parent_collection_lookup=self.collection_lookup,
            collection_item=collection_item,
            collection=sub_collection,
        )

    def test_from_lookup_result_bare_data(self) -> None:
        result = self.make_lookup_result_bare()
        value = DBInputLookupResult.from_lookup_result(result, label="label")
        self.assertEqual(value.label, "label")
        self.assertEqual(value.result_type, LookupResultType.BARE)
        self.assertEqual(value.parent_collection_lookup, self.collection_lookup)
        self.assertEqual(value.collection_item_id, result.collection_item.id)
        self.assertIsNone(value.artifact_id)
        self.assertIsNone(value.collection_id)
        with self.assertNumQueries(0):
            self.assertEqual(value.result, result)

    def test_from_lookup_result_artifact(self) -> None:
        result = self.make_lookup_result_artifact()
        assert result.collection_item is not None
        value = DBInputLookupResult.from_lookup_result(result, label="label")
        self.assertEqual(value.label, "label")
        self.assertEqual(value.result_type, LookupResultType.ARTIFACT)
        self.assertEqual(value.parent_collection_lookup, self.collection_lookup)
        self.assertEqual(value.collection_item_id, result.collection_item.id)
        self.assertEqual(value.artifact_id, result.artifact.id)
        self.assertIsNone(value.collection_id)
        with self.assertNumQueries(0):
            self.assertEqual(value.result, result)

    def test_from_lookup_result_collection(self) -> None:
        result = self.make_lookup_result_collection()
        assert result.collection_item is not None
        value = DBInputLookupResult.from_lookup_result(result, label="label")
        self.assertEqual(value.label, "label")
        self.assertEqual(value.result_type, LookupResultType.COLLECTION)
        self.assertEqual(value.parent_collection_lookup, self.collection_lookup)
        self.assertEqual(value.collection_item_id, result.collection_item.id)
        self.assertIsNone(value.artifact_id)
        self.assertEqual(value.collection_id, result.collection.id)
        with self.assertNumQueries(0):
            self.assertEqual(value.result, result)

    def test_cache_miss_bare_data(self) -> None:
        result = self.make_lookup_result_bare()
        value = DBInputLookupResult.from_lookup_result(result, label="label")
        value._result = None
        with self.assertNumQueries(1):
            self.assertEqual(value.result, result)

    def test_cache_miss_artifact(self) -> None:
        result = self.make_lookup_result_artifact()
        value = DBInputLookupResult.from_lookup_result(result, label="label")
        value._result = None
        with self.assertNumQueries(2):
            self.assertEqual(value.result, result)

    def test_cache_miss_collection(self) -> None:
        result = self.make_lookup_result_collection()
        value = DBInputLookupResult.from_lookup_result(result, label="label")
        value._result = None
        with self.assertNumQueries(2):
            self.assertEqual(value.result, result)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Django ORM utilities."""

from django.db.models import BooleanField, CharField, Func, Value
from django.db.models.expressions import Combinable, CombinedExpression


class JSONBEach(Func):
    """
    Expand a JSON object into a set of key/value pairs.

    Since Django doesn't support set-returning functions very well, this is
    just a minimal wrapper and in practice needs to be used with ``RawSQL``.
    """

    function = "jsonb_each"
    arity = 1


class JSONBEachText(Func):
    """
    Expand a JSON object into a set of key/value pairs, parsing values as text.

    Since Django doesn't support set-returning functions very well, this is
    just a minimal wrapper and in practice needs to be used with ``RawSQL``.
    """

    function = "jsonb_each_text"
    arity = 1


class JSONBToRecordSet(Func):
    """
    Expand a JSON array into a set of its rows.

    Since Django doesn't support set-returning functions very well, this is
    just a minimal wrapper and in practice needs to be used with ``RawSQL``.
    The join must explicitly define the row structure using the table name
    and ``use_alias=False``.
    """

    function = "jsonb_to_recordset"
    arity = 1


class JSONBTypeOf(Func):
    """Return the type of a JSON value."""

    function = "jsonb_typeof"
    arity = 1
    output_field = CharField()


class JSONStringValue(CombinedExpression):
    """Parse a JSON value already known to be a string."""

    output_field = CharField()

    def __init__(self, item: Combinable) -> None:
        """Construct the expression."""
        super().__init__(item, "->>", Value(0))


class Encode(Func):
    """
    Encode binary data into a text representation.

    Note that this follows PostgreSQL's naming for encode/decode, which is
    the other way round from Python's (where "encode" goes from text to
    binary).
    """

    function = "encode"
    arity = 2
    output_field = CharField()


class CompareBoolean(CombinedExpression):
    """Compare two expressions, returning a boolean."""

    output_field = BooleanField()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for VirtualHostMiddleware."""

from django.conf import settings
from django.http import HttpResponse
from django.test import RequestFactory, override_settings

from debusine.server.middlewares.virtual_hosts import VirtualHostMiddleware
from debusine.test.django import TestCase


@override_settings(ALLOWED_HOSTS=["*"])
class VirtualHostMiddlewareTests(TestCase):
    """Test VirtualHostMiddleware."""

    def test_debian_archive_fqdn(self) -> None:
        wrapped_view = VirtualHostMiddleware(lambda _: HttpResponse())
        request = RequestFactory().get(
            "/", HTTP_HOST=settings.DEBUSINE_DEBIAN_ARCHIVE_FQDN
        )
        wrapped_view(request)
        self.assertEqual(
            getattr(request, "urlconf"), "debusine.web.archives.urls"
        )

    @override_settings(
        DEBUSINE_DEBIAN_ARCHIVE_FQDN=["deb.example.com", "deb.example.org"]
    )
    def test_any_of_debian_archive_fqdn(self) -> None:
        wrapped_view = VirtualHostMiddleware(lambda _: HttpResponse())
        request = RequestFactory().get("/", HTTP_HOST="deb.example.org")
        wrapped_view(request)
        self.assertEqual(
            getattr(request, "urlconf"), "debusine.web.archives.urls"
        )

    def test_other_fqdn(self) -> None:
        wrapped_view = VirtualHostMiddleware(lambda _: HttpResponse())
        request = RequestFactory().get("/", HTTP_HOST=settings.DEBUSINE_FQDN)
        wrapped_view(request)
        self.assertIsNone(getattr(request, "urlconf", None))

    @override_settings(
        DEBUSINE_DEBIAN_ARCHIVE_FQDN=["deb.example.com", "deb.example.org"]
    )
    def test_not_any_of_debian_archive_fqdn(self) -> None:
        wrapped_view = VirtualHostMiddleware(lambda _: HttpResponse())
        request = RequestFactory().get("/", HTTP_HOST="deb.example.net")
        wrapped_view(request)
        self.assertIsNone(getattr(request, "urlconf", None))



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""no operation task. Waits and returns. Used in the integration tests."""

from pathlib import Path
from typing import override

from debusine.artifacts.models import WorkRequestResults
from debusine.tasks import BaseExternalTask
from debusine.tasks.models import NoopData


class Noop(BaseExternalTask[NoopData]):
    """
    Task that returns a boolean (execute() depending on the result field).

    Used for integration testing.
    """

    TASK_VERSION = 1

    @override
    def get_subject(self) -> str | None:
        return None

    @override
    def run(self, execute_directory: Path) -> WorkRequestResults:  # noqa: U100
        """Return self.data.result (was sent by the client)."""
        return self.data.result



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Signing task to sign an upload using `debsign`.

This is separate from the `Sign` task because signing uploads is a
customized operation involving signing multiple files and possibly updating
checksums in the `.changes` file to match the signed versions of other
files.
"""

import os
import shutil
from pathlib import Path
from typing import Any, override
from urllib.parse import urljoin

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine import utils
from debusine.artifacts import Upload
from debusine.artifacts.models import ArtifactCategory, WorkRequestResults
from debusine.assets import AssetCategory, KeyPurpose
from debusine.client.models import (
    AssetPermissionCheckResponse,
    FilesResponseType,
    RelationType,
)
from debusine.signing.db.models import Key, sign
from debusine.signing.models import SigningMode
from debusine.signing.tasks import BaseSigningTask
from debusine.signing.tasks.models import DebsignData, DebsignDynamicData
from debusine.tasks.inputs import SingleInput
from debusine.utils import read_changes


class Debsign(BaseSigningTask[DebsignData, DebsignDynamicData]):
    """Task that signs an upload using a key and `debsign`."""

    TASK_VERSION = 1

    unsigned = SingleInput()

    @override
    def post_init(self) -> None:
        super().post_init()
        self._files: FilesResponseType | None = None
        self._key: Key | None = None
        self._signed_changes_path: Path | None = None
        self._permission_check: AssetPermissionCheckResponse | None = None

    @override
    @classmethod
    def analyze_worker(cls) -> dict[str, Any]:
        """Report metadata for this task on this worker."""
        metadata = super().analyze_worker()

        metadata[cls.prefix_with_task_name("available")] = (
            utils.is_command_available("debsign")
        )

        return metadata

    @override
    @classmethod
    def get_provided_worker_tags(cls) -> set[str]:
        tags = super().get_provided_worker_tags()
        if utils.is_command_available("debsign"):
            tags.add(wtags.CAP_PREFIX + "debsign")
        return tags

    @override
    def compute_scheduler_tags(
        self,
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        provided, required = super().compute_scheduler_tags()

        required.add(
            wtags.ProvenanceRequired.SYSTEM, {wtags.CAP_PREFIX + "debsign"}
        )

        return provided, required

    @override
    def can_run_on(self, worker_metadata: dict[str, Any]) -> bool:
        """Check if the specified worker can run the task."""
        if not super().can_run_on(worker_metadata):
            return False

        available_key = self.prefix_with_task_name("available")
        if not worker_metadata.get(available_key, False):
            return False

        return True

    @override
    def get_subject(self) -> str | None:
        return None

    @override
    def compute_dynamic_data(self) -> DebsignDynamicData:
        """Resolve artifact lookups for this task."""
        return DebsignDynamicData(unsigned_id=self.unsigned.artifact_id)

    @override
    def fetch_input(self, destination: Path) -> bool:
        """Fetch artifacts and database objects needed by the task."""
        assert self.debusine is not None
        assert self.dynamic_data is not None
        assert self.work_request_id is not None
        assert self.workspace_name is not None

        unsigned = self.fetch_artifact(
            self.dynamic_data.unsigned_id, destination
        )
        if unsigned.category != ArtifactCategory.UPLOAD:
            self.append_to_log_file(
                "fetch_input.log",
                [
                    f"Expected unsigned artifact of category "
                    f"{ArtifactCategory.UPLOAD}; got {unsigned.category}"
                ],
            )
            return False
        self._files = unsigned.files

        asset_slug = f"openpgp:{self.data.key}"
        permission_check = self.debusine.asset_permission_check(
            asset_category=AssetCategory.SIGNING_KEY,
            asset_slug=asset_slug,
            permission_name="sign_with",
            artifact_id=unsigned.id,
            work_request_id=self.work_request_id,
            workspace=self.workspace_name,
        )
        self._permission_check = permission_check
        if not permission_check.has_permission:
            self.append_to_log_file(
                "fetch_input.log",
                [
                    f"User {permission_check.username} does not have "
                    f"permission to sign {unsigned.id} with {asset_slug}."
                ],
            )
            return False

        try:
            self._key = Key.objects.get(
                purpose=KeyPurpose.OPENPGP_UPLOAD, fingerprint=self.data.key
            )
        except Key.DoesNotExist:
            self.append_to_log_file(
                "fetch_input.log",
                [f"Signing key openpgp:{self.data.key} does not exist"],
            )
            return False

        return True

    @override
    def configure_for_execution(
        self,
        download_directory: Path,  # noqa: U100
    ) -> bool:
        """Configure task variables."""
        # Nothing to do here; fetch_input does it all.
        return True

    @override
    def prepare_to_run(
        self, download_directory: Path, execute_directory: Path
    ) -> None:
        """Copy downloaded files into the execution directory."""
        # download_directory and execute_directory are always on the same
        # file system.
        shutil.copytree(
            download_directory,
            execute_directory / "input",
            copy_function=os.link,
        )
        self.update_changes_file(execute_directory / "input")

    def _changes_file(self) -> str:
        """Locate the .changes file in _.files."""
        assert self._files is not None

        # Upload.files_contain_changes checks that there is exactly one
        # .changes file.
        [changes_file] = [
            file for file in self._files if file.endswith(".changes")
        ]
        return changes_file

    def update_changes_file(self, input_directory: Path) -> None:
        """Record the debusine uploader in the .changes file."""
        assert self.debusine is not None
        assert self.work_request_id is not None

        changes_file = input_directory / self._changes_file()
        changes = read_changes(changes_file)
        work_request = self.debusine.work_request_get(self.work_request_id)

        base_url = self.debusine.base_api_url
        for changes_key, link in (
            ("Debusine-User", "webui_created_by"),
            ("Debusine-Workflow", "webui_workflow_root"),
            ("Debusine-Work-Request", "webui_self"),
        ):
            if link in work_request.links:
                changes[changes_key] = urljoin(
                    base_url, work_request.get_link_url(link)
                )
            else:
                changes.pop(changes_key, None)
        with changes_file.open("w") as f:
            changes.dump(f, text_mode=True)

    @override
    def run(self, execute_directory: Path) -> WorkRequestResults:
        """Sign all the requested files."""
        assert self.work_request_id is not None
        assert self._key is not None
        assert self._permission_check is not None

        with self.open_debug_log_file("cmd-output.log", mode="wb") as log_file:
            changes_file = self._changes_file()
            data_path = execute_directory / "input" / changes_file
            signature_path = execute_directory / "output" / changes_file
            signature_path.parent.mkdir(parents=True, exist_ok=True)
            try:
                sign(
                    [self._key],
                    data_path,
                    signature_path,
                    SigningMode.DEBSIGN,
                    self.work_request_id,
                    username=self._permission_check.username,
                    user_id=self._permission_check.user_id,
                    resource=self._permission_check.resource,
                    log_file=log_file,
                )
            except Exception as e:
                log_file.write(f"Debsign failed: {e}\n".encode())
                raise
            else:
                self._signed_changes_path = signature_path

        return WorkRequestResults.SUCCESS

    @override
    def upload_artifacts(
        self,
        execute_directory: Path,
        *,
        execution_result: WorkRequestResults,
    ) -> None:
        """Upload artifacts for the task."""
        assert self.dynamic_data is not None
        assert self.work_request_id is not None
        assert self.debusine is not None
        assert self._signed_changes_path is not None
        # run() always either returns SUCCESS or raises an exception; in the
        # latter case we won't get here.
        assert execution_result == WorkRequestResults.SUCCESS

        upload_artifact = Upload.create(changes_file=self._signed_changes_path)
        uploaded_upload_artifact = self.debusine.upload_artifact(
            upload_artifact,
            workspace=self.workspace_name,
            work_request=self.work_request_id,
        )
        self.debusine.relation_create(
            uploaded_upload_artifact.id,
            self.dynamic_data.unsigned_id,
            RelationType.RELATES_TO,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for functions in debusine.project_utils."""

from pathlib import Path

from django.core.exceptions import ImproperlyConfigured

from debusine.project import project_utils
from debusine.test.django import TestCase


class ProjectUtilsTests(TestCase):
    """Tests for functions in defaults.py."""

    SECRET_KEY = "this-is-a-secret-key"

    def create_secret_key_file(self, permissions: int) -> Path:
        """Create a secret key file, schedules deletion after test."""
        file = self.create_temporary_file(
            contents=self.SECRET_KEY.encode("utf-8")
        )
        file.chmod(permissions)
        return file

    def test_read_secret_key_reads_key(self) -> None:
        """read_secret_key return the key in the file."""
        secret_key_file = self.create_secret_key_file(0o600)
        self.assertEqual(
            self.SECRET_KEY, project_utils.read_secret_key(secret_key_file)
        )

    def test_read_secret_key_cannot_read_file(self) -> None:
        """read_secret_key cannot read the file, raises ImproperlyConfigured."""
        with self.assertRaisesRegex(ImproperlyConfigured, "Cannot read"):
            project_utils.read_secret_key(
                "/tmp/a-file-that-does-not-exist-24241"
            )

    def test_read_secret_key_file_permissions_too_open(self) -> None:
        """read_secret_key file permissions too open, raises exception."""
        secret_key_file = self.create_secret_key_file(0o640)

        with self.assertRaisesRegex(ImproperlyConfigured, "too open"):
            project_utils.read_secret_key(secret_key_file)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Functions for rendering input fields."""

from collections.abc import Generator
from typing import Any, override

from django.template.context import Context
from django.utils.html import format_html
from django.utils.safestring import SafeString

import debusine.db.tasks.inputs as dbinputs
from debusine.db.models import Artifact, Collection
from debusine.db.tasks import DBTask
from debusine.tasks.inputs import inputs
from debusine.tasks.models import (
    BaseTaskInputArtifact,
    ExtraExternalRepository,
    InputArtifactMultiple,
    InputArtifactSingle,
    InputCollectionSingle,
    LookupMultiple,
    LookupSingle,
)
from debusine.web.views.base import TemplateWidget, Widget
from debusine.web.views.view_utils import highlight_as_yaml


def render_artifact(
    value: BaseTaskInputArtifact, context: Context
) -> SafeString:
    """Render a BaseInputArtifactSingle."""
    # Try to access a cached version of the artifact
    artifact: Artifact | None
    if (artifact := getattr(value, "_artifact", None)) is None:
        # If it fails, try to look it up
        try:
            artifact = Artifact.objects.get(pk=value.artifact_id)
            if hasattr(value, "_artifact"):
                value._artifact = artifact
        except Artifact.DoesNotExist:
            artifact = None

    assert context.template is not None
    template = context.template.engine.get_template(
        "web/_input-field-row-artifact.html"
    )
    with context.update({"value": value, "artifact": artifact}):
        return template.render(context)


def render_collection(
    value: InputCollectionSingle, context: Context
) -> SafeString:
    """Render an InputCollectionSingle."""
    # Try to access a cached version of the collection
    collection: Collection | None
    if (collection := getattr(value, "_collection", None)) is None:
        # If it fails, try to look it up
        try:
            collection = Collection.objects.get(pk=value.collection_id)
            if hasattr(value, "_collection"):
                value._collection = collection
        except Collection.DoesNotExist:
            collection = None

    assert context.template is not None
    template = context.template.engine.get_template(
        "web/_input-field-row-collection.html"
    )
    with context.update({"value": value, "collection": collection}):
        return template.render(context)


def render_result_label(label: str) -> SafeString:
    """Render a label from a lookup result."""
    return format_html("<code>{label}</code>", label=label)


def render_lookup(lookup: LookupSingle | LookupMultiple) -> SafeString:
    """Render a lookup."""
    match lookup:
        case int() | str():
            return format_html("<code>{value}</code>", value=lookup)
        case LookupMultiple():
            return highlight_as_yaml(lookup.export())
        case _:
            return highlight_as_yaml(lookup)


class InputFieldWidget[ValueType](Widget):
    """Base for widgets rendering the value of an input field."""

    def __init__(self, field: inputs.TaskInput[Any], value: ValueType) -> None:
        """
        Initialize common members.

        :param field: the input field
        :param value: the current value of the input field
        """
        self.name = field.name
        self.verbose_name = (
            field.verbose_name or field.name.replace("_", " ").capitalize()
        )
        self.help_text = field.help_text
        self.field = field
        self.value = value

    def render_details(self) -> SafeString:
        """Get the text for the tooltip on the field name."""
        result = SafeString()

        if self.help_text and (help_text := self.help_text.rstrip()):
            if help_text[-1] != ".":
                help_text += "."
            result += format_html(
                """<p class="mb-3 fst-italic">{description}</p>""",
                description=help_text,
            )

        result += SafeString("<dl>")
        for label, content in self.render_details_rows():
            result += format_html(
                "<dt>{label}:</dt><dd>{content}</dd>",
                label=label,
                content=content,
            )
        result += SafeString("</dl>")
        return result

    def render_row_title(self, context: Context) -> SafeString:
        """Render the title cell of the result row."""
        assert context.template is not None
        template = context.template.engine.get_template(
            "web/_input-field-row-title.html"
        )
        with context.update({"widget": self}):
            return template.render(context)

    def render_details_rows(self) -> Generator[tuple[SafeString, SafeString]]:
        """Render term, description pairs for the details DL element."""
        yield (
            SafeString("Defined as"),
            format_html("<code>{name}</code>", name=self.name),
        )

    def render_value_rows(self, _: Context) -> Generator[SafeString]:
        """Render the rows for the field value."""
        serialized = self.field.serialize_value(self.value)
        yield highlight_as_yaml(serialized)

    @override
    def render(self, context: Context) -> str:
        if not (rows := list(self.render_value_rows(context))):
            return SafeString()
        with context.update({"rows": rows}):
            result = format_html(
                "<tr>{title}<td>{row}</td></tr>",
                title=self.render_row_title(context),
                row=rows[0],
            )
            for row in rows[1:]:
                result += format_html("<tr><td>{row}</td></tr>", row=row)
        return result

    @classmethod
    def create(
        cls, container: inputs.TaskFieldContainer, field: inputs.TaskInput[Any]
    ) -> "InputFieldWidget[Any]":
        """Get the widget for an input field."""
        if not container.field_is_resolved(field.name):
            return UnresolvedFieldWidget(field, None)

        value = field.get_value(container)
        match field:
            case (
                inputs.SingleInput()
                | inputs.OptionalSingleInput()
                | inputs.EnvironmentInput()
                | inputs.OptionalEnvironmentInput()
                | dbinputs.ArtifactInput()
            ):
                return ArtifactInputFieldWidget(field, value)

            case inputs.SingleInputList():
                return ArtifactInputListFieldWidget(field, value)

            case inputs.MultiInput() | inputs.UploadArtifactsInput():
                return InputArtifactMultipleFieldWidget(field, value)

            # TODO: list[InputArtifactMultiple]
            # case inputs.MultiInputList():
            #    return InputFieldWidget(field, value)

            # TODO: tuple[InputCollectionSingle, InputCollectionSingle | None]
            # case inputs.SuiteArchiveInput():
            #    return InputFieldWidget(field, value)

            case inputs.ExtraRepositoriesInput():
                return ExtraExternalRepositoryFieldWidget(field, value)

            case inputs.DebusineFQDNInput() | inputs.BuildArchitectureInput():
                return CodeStringInputFieldWidget(field, value)

            case (
                dbinputs.CollectionInput()
                | dbinputs.TaskConfigurationCollection()
            ):
                return CollectionInputFieldWidget(field, value)

            case _:
                return InputFieldWidget(field, value)


class CodeStringInputFieldWidget(InputFieldWidget[str | None]):
    """Widget for string values rendered inside ``<code>`` tags."""

    @override
    def render_value_rows(self, context: Context) -> Generator[SafeString]:
        yield format_html("<code>{value}</code>", value=self.value)


class ArtifactInputFieldWidget(InputFieldWidget[InputArtifactSingle | None]):
    """Widget for InputArtifactSingle values."""

    @override
    def render_details_rows(self) -> Generator[tuple[SafeString, SafeString]]:
        yield from super().render_details_rows()
        if self.value is None:
            return
        yield (
            SafeString("Task input field"),
            render_result_label(self.value.label),
        )
        yield SafeString("Lookup"), render_lookup(self.value.lookup)
        # TODO: highlight value.data as YAML in a tooltip or popup?

    @override
    def render_value_rows(self, context: Context) -> Generator[SafeString]:
        if self.value is None:
            return
        yield render_artifact(self.value, context)


class ArtifactInputListFieldWidget(InputFieldWidget[list[InputArtifactSingle]]):
    """Widget for list[InputArtifactSingle] values."""

    @override
    def render_details_rows(self) -> Generator[tuple[SafeString, SafeString]]:
        yield from super().render_details_rows()
        if not self.value:
            return
        for value in self.value:
            yield (
                format_html(
                    "Task input field {label}",
                    label=render_result_label(value.label),
                ),
                render_lookup(value.lookup),
            )

    @override
    def render_value_rows(self, context: Context) -> Generator[SafeString]:
        if not self.value:
            return
        for value in self.value:
            yield render_artifact(value, context)


class InputArtifactMultipleFieldWidget(InputFieldWidget[InputArtifactMultiple]):
    """Widget for InputArtifactMultiple values."""

    @override
    def render_details_rows(self) -> Generator[tuple[SafeString, SafeString]]:
        yield from super().render_details_rows()
        yield (
            SafeString("Task input field"),
            render_result_label(self.value.label),
        )
        yield SafeString("Lookup"), render_lookup(self.value.lookup)

    @override
    def render_value_rows(self, context: Context) -> Generator[SafeString]:
        if not self.value.artifacts:
            return
        for value in self.value.artifacts:
            yield render_artifact(value, context)


class CollectionInputFieldWidget(
    InputFieldWidget[InputCollectionSingle | None]
):
    """Widget for InputCollectionSingle values."""

    @override
    def render_details_rows(self) -> Generator[tuple[SafeString, SafeString]]:
        yield from super().render_details_rows()
        if self.value is None:
            return
        yield (
            SafeString("Task input field"),
            render_result_label(self.value.label),
        )
        yield SafeString("Lookup"), render_lookup(self.value.lookup)
        # TODO: highlight value.data as YAML in a tooltip or popup?

    @override
    def render_value_rows(self, context: Context) -> Generator[SafeString]:
        if self.value is None:
            return
        yield render_collection(self.value, context)


class ExtraExternalRepositoryFieldWidget(
    InputFieldWidget[list[ExtraExternalRepository] | None]
):
    """Widget for ExtraExternalRepositoryFieldWidget values."""

    @override
    def render_value_rows(self, context: Context) -> Generator[SafeString]:
        if self.value is None:
            return
        for repo in self.value:
            yield format_html(
                """
                <details>
                <summary>deb {url} {suite} {components}</summary>
                <pre>{repo}</pre>
                </details>
                """,
                url=repo.url,
                suite=repo.suite,
                components=" ".join(repo.components) if repo.components else "",
                repo=repo.as_deb822_source(),
            )


class UnresolvedFieldWidget(InputFieldWidget[None]):
    """Widget for unresolved fields."""

    @override
    def render_value_rows(self, context: Context) -> Generator[SafeString]:
        yield SafeString("unresolved")


class TaskInputsWidget(TemplateWidget):
    """Render input fields for a task."""

    template_name: str = "web/_input-fields.html"

    def __init__(
        self,
        task: DBTask[Any, Any],
        element_id: str = "task-input-fields",
        internal: bool = False,
    ) -> None:
        """
        Initialize a TaskInputsWidget.

        :param task: task with the input fields to render
        :param class_name: class name of the toplevel element
        """
        self.task = task
        self.element_id = element_id
        self.internal = internal
        self.fields: list[InputFieldWidget[Any]] = []
        for cont, field in sorted(
            self.task.iter_all_fields(), key=lambda x: x[1].name
        ):
            if field.hidden:
                continue
            if self.internal != field.internal:
                continue
            self.fields.append(InputFieldWidget.create(cont, field))

    def __bool__(self) -> bool:
        """Check if there are input fields to display."""
        return bool(self.fields)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the UIShortcuts base view."""

from django.template import Context
from django.utils.safestring import SafeString

from debusine.db.models import Artifact, User, WorkRequest
from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.icons import Icons
from debusine.web.views import sidebar
from debusine.web.views.base import Widget
from debusine.web.views.sidebar import SidebarAction, SidebarElement


class TestSidebar(TestCase):
    """Tests for Sidebar."""

    scenario = scenarios.DefaultContext()

    def _render(self, widget: Widget) -> str | SafeString:
        """Render a widget."""
        return widget.render(Context())

    def test_info(self) -> None:
        """Test selection of information for UI."""
        kwargs = {"icon": "I", "label": "L"}

        el = SidebarElement(label="L")
        self.assertEqual(el.icon, "app")
        self.assertEqual(el.content, "-")
        self.assertEqual(el.main_tooltip, "L")
        self.assertIsNone(el.icon_tooltip)

        el = SidebarElement(**kwargs)
        self.assertEqual(el.icon, "I")
        self.assertEqual(el.content, "-")
        self.assertEqual(el.main_tooltip, "L")
        self.assertIsNone(el.icon_tooltip)

        el = SidebarElement(value="V", **kwargs)
        self.assertEqual(el.icon, "I")
        self.assertEqual(el.content, "V")
        self.assertEqual(el.main_tooltip, "L")
        self.assertIsNone(el.icon_tooltip)

        el = SidebarElement(detail="D", **kwargs)
        self.assertEqual(el.icon, "I")
        self.assertEqual(el.content, "D")
        self.assertEqual(el.main_tooltip, "L")
        self.assertIsNone(el.icon_tooltip)

        el = SidebarElement(value="V", detail="D", **kwargs)
        self.assertEqual(el.icon, "I")
        self.assertEqual(el.content, "V")
        self.assertEqual(el.main_tooltip, "D")
        self.assertEqual(el.icon_tooltip, "L")

    def test_render_action(self) -> None:
        """Test SidebarAction rendering."""
        kwargs = {"icon": "I", "label": "L", "url": "U"}
        classes = (
            '"list-group-item list-group-item-action'
            ' d-flex justify-content-between"'
        )

        el = SidebarAction(detail="D", value="V", **kwargs)
        self.assertEqual(
            self._render(el),
            f'<a class={classes} title="D" href="U">'
            '<span><span class="bi bi-I" title="L"></span> V</span>'
            '<span>…</span></a>',
        )

        el = SidebarAction(detail="D", **kwargs)
        self.assertEqual(
            self._render(el),
            f'<a class={classes} title="L" href="U">'
            '<span><span class="bi bi-I"></span> D</span>'
            '<span>…</span></a>',
        )

        el = SidebarAction(value="V", **kwargs)
        self.assertEqual(
            self._render(el),
            f'<a class={classes} title="L" href="U">'
            '<span><span class="bi bi-I"></span> V</span>'
            '<span>…</span></a>',
        )

        el = SidebarAction(**kwargs)
        self.assertEqual(
            self._render(el),
            f'<a class={classes} title="L" href="U">'
            '<span><span class="bi bi-I"></span> -</span>'
            '<span>…</span></a>',
        )

    def test_render_element(self) -> None:
        """Test SidebarElement rendering."""
        kwargs = {"icon": "I", "label": "L"}

        el = SidebarElement(detail="D", value="V", **kwargs)
        self.assertEqual(
            self._render(el),
            '<span class="list-group-item" title="D">'
            '<span class="bi bi-I" title="L"></span> V</span>',
        )

        el = SidebarElement(detail="D", **kwargs)
        self.assertEqual(
            self._render(el),
            '<span class="list-group-item" title="L">'
            '<span class="bi bi-I"></span> D</span>',
        )

        el = SidebarElement(value="V", **kwargs)
        self.assertEqual(
            self._render(el),
            '<span class="list-group-item" title="L">'
            '<span class="bi bi-I"></span> V</span>',
        )

        el = SidebarElement(**kwargs)
        self.assertEqual(
            self._render(el),
            '<span class="list-group-item" title="L">'
            '<span class="bi bi-I"></span> -</span>',
        )

    def test_user(self) -> None:
        """Test create_user."""
        artifact = Artifact()
        work_request = WorkRequest(status=WorkRequest.Statuses.PENDING)
        user = User(username="test")

        el = sidebar.create_user(user)
        self.assertEqual(el.label, "User")
        self.assertEqual(el.value, "test")

        el = sidebar.create_user(user, context=artifact)
        self.assertEqual(el.label, "User who created the artifact")
        self.assertEqual(el.value, "test")

        el = sidebar.create_user(user, context=work_request)
        self.assertEqual(el.label, "User who created the work request")
        self.assertEqual(el.value, "test")


class TestCreateWorker(TestCase):
    """Test :py:func:`create_worker`."""

    def test_none(self) -> None:
        """Test with no worker."""
        item = sidebar.create_worker(None)
        self.assertEqual(item.label, "Worker")
        self.assertEqual(item.icon, Icons.WORKER)
        self.assertIsNone(item.value)
        self.assertFalse(hasattr(item, "url"))

    def test_worker(self) -> None:
        """Test with a worker."""
        worker = self.playground.create_worker()
        item = sidebar.create_worker(worker)
        assert isinstance(item, SidebarAction)
        self.assertEqual(item.label, "Worker")
        self.assertEqual(item.icon, Icons.WORKER)
        self.assertEqual(item.value, worker.name)
        self.assertEqual(item.url, worker.get_absolute_url())
